from typing import *


# =====================================================================================================================
class GetattrClsEcho_Meta(type):
    """
    GOAL
    ----
    for any not existed attribute return attr-name!

    CREATED SPECIALLY FOR
    ---------------------
    here is GetattrClsEcho
    """
    def __getattr__(cls, item: str) -> str:
        """if no exists attr/meth
        """
        if getattr(cls, "_UNDERSCORE_AS_SPACE"):
            item = item.replace("_", " ")
        return item


# =====================================================================================================================
class GetattrClsEcho(metaclass=GetattrClsEcho_Meta):
    """
    GOAL
    ----
    just use class as string values over attributes.
    If you dont want to keep original strings in code.
    just to see maybe it will be pretty convenient.

    CREATED SPECIALLY FOR
    ---------------------
    everyday usage

    NOTE
    ----
    of cause you cant apply any chars (like punctuation) here except Literals cause of name constraints.

    WHY NOT: just using direct strings?
    -----------------------------------

    BEST USAGE
    ----------
    assert GetattrClsEcho.hello == "hello"
    assert GetattrClsEcho.hello_world == "hello_world"
    print(GetattrClsEcho.Hello)   # "Hello"
    """
    _UNDERSCORE_AS_SPACE: bool | None = None


class GetattrClsEchoSpace(GetattrClsEcho):
    """
    GOAL
    ----
    SAME AS: base parent class GetattrClsEcho! see all there!
    DIFFERENCE: just replaced all UNDERSCORE-signs by Space!
    """
    _UNDERSCORE_AS_SPACE = True


# =====================================================================================================================



from base_aux.base_inits.m2_nest_init_source_kwargs import *
from base_aux.aux_types.m1_type_aux import *
from base_aux.base_statics.m1_types import *
from base_aux.aux_attr.m1_attr1_aux import *


# =====================================================================================================================
@final
class EqAux(NestInit_Source):
    # FIXME: decide about callables/KwAgs by now NOT USE IT! just direct cmp eq!
    # CALLABLES: bool = None   # NOTE: DONT use callables here in Eq!!!
    # NOTE: dont use Args/Kwargs here in EqAux! - no callables! just final objects!

    # -----------------------------------------------------------------------------------------------------------------
    def check_oneside__exx(self, other: Any, return_bool: bool = None) -> bool | Exception:
        # if self.CALLABLES:
        #     self.SOURCE = CallableAux(self.SOURCE).resolve_exx()
        #
        #     try:
        #         other = CallableAux(other).resolve_raise()
        #     except Exception as exx:
        #         return exx

        # EXX ------
        if TypeAux(other).check__exception():
            if TypeAux(self.SOURCE).check__nested__by_cls_or_inst(other):   # CORRECT ORDER!!!
                return True
        # WORK ------
        try:
            result = self.SOURCE == other
            if result:
                return True
        except Exception as exx:
            result = exx
            # if TypeAux(other).check__exception() and TypeAux(result12).check__nested__by_cls_or_inst(other):
            #     return True
            if return_bool:
                result = False

        # FINAL ------
        return result

    def check_oneside__bool(self, other: Any) -> bool:
        return self.check_oneside__exx(other, return_bool=True)

    def check_oneside__reverse(self, other: Any) -> bool:
        return self.check_oneside__bool(other) is not True

    # -----------------------------------------------------------------------------------------------------------------
    def check_doubleside__exx(self, other: Any, return_bool: bool = None) -> bool | Exception:
        """
        GOAL
        ----
        just a direct comparing code like
            self.validate_last = self.value_last == self.VALIDATE_LINK or self.VALIDATE_LINK == self.value_last
        will not work correctly

        if any result is True - return True.
        if at least one false - return False
        if both exx - return first exx  # todo: deside return False in here!

        CREATED SPECIALLY FOR
        ---------------------
        manipulate aux_types which have special methods for __cmp__
        for cases when we can switch places

        BEST USAGE
        ----------
            class ClsEq:
                def __init__(self, val):
                    self.VAL = val

                def __eq__(self, other):
                    return other == self.VAL

            assert ClsEq(1) == 1
            assert 1 == ClsEq(1)

            assert compare_doublesided(1, Cls(1)) is True
            assert compare_doublesided(Cls(1), 1) is True

        example above is not clear! cause of comparison works ok if any of object has __eq__() meth even on second place!
        but i think in one case i get ClsException and with switching i get correct result!!! (maybe fake! need explore!)
        """
        # ONESIDE ------
        result12 = self.check_oneside__exx(other=other, return_bool=return_bool)
        if result12 is True:
            return True

        result21 = EqAux(other).check_oneside__exx(other=self.SOURCE, return_bool=return_bool)
        if result21 is True:
            return True

        # BOOL TRUE ------
        if return_bool:
            return False

        # BOOL FALSE ------
        if False in [result12, result21]:
            return False

        # FINAL -----------
        return result12

    def check_doubleside__bool(self, other: Any) -> bool:
        """
        same as compare_doublesided_or_exx but
        in case of ClsException - return False

        CREATED SPECIALLY FOR
        ---------------------
        Valid.value_validate
        """
        return self.check_doubleside__exx(other, return_bool=True)

    def check_doubleside__reverse(self, other: Any) -> bool:
        """
        just reverse result for compare_doublesided__bool
        so never get ClsException, only bool
        """
        return self.check_doubleside__bool(other) is not True

    # -----------------------------------------------------------------------------------------------------------------
    def check_by_dict__direct(self, other: TYPING.KWARGS_FINAL) -> bool:
        """
        GOAL
        ----
        cmp direct values

        CREATED SPECIALLY FOR
        ---------------------
        """
        for key, expected in other.items():
            key_real = AttrAux(self.SOURCE).anycase__name_original(key)
            if key_real is None:
                if expected is None:
                    continue
                else:
                    return False
            try:
                actual = getattr(self.SOURCE, key)
            except Exception as exx:
                actual = exx

            if actual != expected:
                msg = f"for {key_real=} {actual=}/{expected=}"
                print(msg)
                return False


# =====================================================================================================================
# @final
# class EqAuxValidator:
#     # NOTE: for acllables DONT USE ARGS/KWARGS! its just about
#     CALLABLES = True


# =====================================================================================================================
# class EqByAttrs:
#     pass
#     # for dir


# =====================================================================================================================



import time
from typing import *

from PyQt5.QtCore import QThread

import smtplib
import telebot


# =====================================================================================================================
class _AlertInterface:
    """Interface for Alerts

    RULES:
    - if some method cant exists (like for telegram) - just return True!
    - Dont return None!
    - Dont use Try sentences inside - it will be applied in upper logic!
    - Decide inside if it was success or not, and return conclusion True/False only.
    """
    def _connect_unsafe(self) -> Union[bool, NoReturn]:
        """establish connection to source
        """
        return True

    def _login_unsafe(self) -> Union[bool, NoReturn]:
        """authorise to source
        """
        return True

    def _send_unsafe(self) -> Union[bool, NoReturn]:
        """send msg
        """
        return True

    def _msg_compose(self) -> Union[str, 'MIMEMultipart']:
        """generate msg from existed data in attributes (passed before on init)
        """
        pass

    def _recipient_self_get(self) -> str:
        """RECIPIENT SelfSending, get from obvious class aux_types!
        """
        pass


# =====================================================================================================================
class AlertBase(_AlertInterface, QThread):     # REM: DONT ADD SINGLETON!!! SNMP WILL NOT WORK!!! and calling logic will be not simle!
    """Base class for Alert aux_types.

    FEATURE:
    - send msg,
    - threading
        - daemons
        - collect all active threads
        - wait all spawned threads finished

    :ivar SUBJECT_PREFIX: default prefix for subject
    :ivar SUBJECT_NAME: default name for subject
    :ivar body: actual msg body for alert
    :ivar _subtype: it used to change subtype only in smtp (http)

    :ivar BODY_TIMESTAMP_USE: append timestamp into the body
    :ivar TIMESTAMP: initiation timestamp for created instance

    :ivar RECONNECT_LIMIT: how many times it will try to reconnect, after - just close object
    :ivar RECONNECT_PAUSE: pause between reconnecting in seconds

    :ivar RECIPIENT_SPECIAL: recipient for sending msg
        None - if selfSending!

    :ivar _conn: actual connection object
    :ivar _result: result for alert state
        None - in process,
        False - finished UnSuccess,
        True - finished success!

    :ivar _threads_active: spawned (only active) threads
    """
    # SETTINGS ------------------------------------
    SUBJECT_PREFIX: Optional[str] = "[ALERT]"
    SUBJECT_NAME: Optional[str] = None
    body: Union[None, str, dict] = None
    _subtype: Optional[str] = "plain"

    BODY_TIMESTAMP_USE: bool = True
    TIMESTAMP: str = None

    TIMEOUT_SEND: float = 1.2

    RECONNECT_LIMIT: int = 10
    RECONNECT_PAUSE: int = 60
    # TIMEOUT_RATELIMIT: int = 600    # when EXX 451, b'Ratelimit exceeded

    RECIPIENT_SPECIAL: Optional[Any] = None

    # AUX -----------------------------------------
    _conn: Union[None, smtplib.SMTP_SSL, telebot.TeleBot] = None
    _result: Optional[bool] = None

    _threads_active: set['AlertBase'] = set()

    # =================================================================================================================
    def __init__(self, body: Optional[Any] = None, _subj_name: Optional[str] = None, _subtype: Optional[str] = None):
        """Send msg

        :param body: define msg body
            None - just for research! (maybe deprecated!)
        :param _subj_name: reuse new subject name instead of default
        :param _subtype: reuse new _subtype instead of default
        """
        super().__init__()
        self.TIMESTAMP = time.strftime("%Y.%m.%d %H:%M:%S")

        # self._mutex: threading.Lock = threading.Lock()
        self.SUBJECT_NAME = _subj_name or self.SUBJECT_NAME
        self._subtype = _subtype or self._subtype

        # BODY ---------------
        if body is not None:
            body = str(body)
            if self.BODY_TIMESTAMP_USE:
                body += f"\n{self.TIMESTAMP}"
            self.body = body

            self.start()

    # =================================================================================================================
    @property
    def RECIPIENT(self) -> str:
        """RECIPIENT actual/final
        """
        pass
        return self.RECIPIENT_SPECIAL or self._recipient_self_get()

    @property
    def SUBJECT(self) -> str:
        """final msg subject, composed with name and prefix.
        """
        if self.SUBJECT_NAME is None:
            self.SUBJECT_NAME = self.NAME
        return self.SUBJECT_PREFIX + self.SUBJECT_NAME

    @classmethod
    @property
    def NAME(cls) -> str:
        """name for AlertClass.
        Accustom ourselves to name you classes recognisable without adding new one redundant attributes.
        """
        return cls.__name__

    # =================================================================================================================
    def start(self, *args):
        """this is just add ability to collect started threads in class
        """
        self.__class__._threads_active.add(self)
        super().start()

    def _thread_finished(self):
        """del thread object from collection.
        called then thread finished.
        """
        self.__class__._threads_active.discard(self)

    @classmethod
    def threads_wait_all(cls):
        """wait while all spawned active threads will finished.
        """
        try:
            time.sleep(1)
            while cls._threads_active:
                list(cls._threads_active)[0].wait()
        except:
            pass

    def result_wait(self) -> Optional[bool]:
        """wait for finish thread and get succession result.
        Created for tests mainly! but you can use!
        """
        self.wait()
        return self._result

    # =================================================================================================================
    def _conn__check_exists(self) -> bool:
        """check if connection object exists
        """
        return self._conn is not None

    def _conn__disconnect(self) -> None:
        """disconnect connection object
        """
        if self._conn:
            self._conn.quit()
        self._conn__clear()

    def _conn__clear(self) -> None:
        """del connection object
        """
        self._conn = None

    def _connect(self) -> Optional[bool]:
        """create connection object
        """
        result = None
        if not self._conn__check_exists():
            print(f"TRY _connect {self.__class__.__name__}")
            try:
                if self._connect_unsafe():
                    print("[READY] connect")

            except Exception as exx:
                print(f"[CRITICAL] CONNECT [{exx!r}]")
                self._conn__clear()

        if self._conn__check_exists():
            try:
                result = self._login_unsafe()
                if result:
                    print("[READY] login")
            except Exception as exx:
                print(f"[CRITICAL] LOGIN [{exx!r}]")
                self._conn__clear()

        print("="*100)
        print("="*100)
        print("="*100)
        print()

        return result

    # =================================================================================================================
    def run(self) -> None:
        """main logic which manage started thread
        """
        counter = 0
        while not self._conn__check_exists() and counter <= self.RECONNECT_LIMIT:
            counter += 1
            if not self._connect():
                print(f"[WARNING]try [{counter=}]")
                print("=" * 100)
                print()
                time.sleep(self.RECONNECT_PAUSE)

        print("[Try send", "-" * 80)
        print(self._msg_compose())
        print("Try send]", "-" * 80)

        if self._conn__check_exists():
            try:
                result = self._send_unsafe()
                if result:
                    print("[READY] send")
                    self._result = True
            except Exception as exx:
                msg = f"[CRITICAL]unexpected [{exx!r}]"
                print(msg)
                self._conn__clear()

        print()
        print()
        print()

        if self._result is None:
            self._result = False

        self._thread_finished()


# =====================================================================================================================



class PROJECT:
    # PROJECT ----------------------------------------------
    NAME_IMPORT: str = "pyqt_templates"
    KEYWORDS: list[str] = [
        "pyqt",
        "pyqt templates", "pyqt guide",
        "pyqt usage", "pyqt examples", "pyqt usage examples", "pyqt help", "pyqt help examples",
        "pyqt signals",
        "pyqt highlighter",
    ]
    # README --------------------------------------------------
    # add DOUBLE SPACE at the end of all lines! for correct representation in MD-viewers
    DESCRIPTION_SHORT: str = "pyqt help examples and some other useful aux_types (overloaded pyqt classes)"
    DESCRIPTION_LONG: str = """Designed for ..."""
    FEATURES: list[str] = [
        # "feat1",
        # ["feat2", "block1", "block2"],

        "good template for TableView/Model/Signals",
    ]

    # HISTORY -----------------------------------------------
    VERSION: tuple[int, int, int] = (0, 1, 6)
    TODO: list[str] = [
        "add Events for TM/TV/PTE/...",
    ]
    FIXME: list[str] = [
        "..."
    ]
    NEWS: list[str] = [
        "[annotAttrs]apply new ver",
    ]


# =====================================================================================================================



from typing import *
import json
import re
import string

from base_aux.base_statics.m1_types import *
from base_aux.aux_text.m0_patterns import *
from base_aux.aux_text.m4_ini import ConfigParserMod


# =====================================================================================================================
# @final      # dont use final here! expect nesting for fileWork! or FIXME: nest FileAux here!????
class TextAux:
    TEXT: TYPING.STR_FINAL = ""

    def __init__(self, text: TYPING.STR_DRAFT = "", *args, **kwargs) -> None | NoReturn:
        if text is not None:
            self.TEXT = str(text)
        super().__init__(*args, **kwargs)

    # =================================================================================================================
    def sub__regexp(self, pat: str, new: str | None = None, flags: re.RegexFlag = 0, *, cover_type: PatCoverStyle = PatCoverStyle.NONE) -> str:
        if new is None:
            new = ""

        flags = flags or 0

        if cover_type == PatCoverStyle.WORD:
            pat = r"\b" + pat + r"\b"

        elif cover_type == PatCoverStyle.LINE:
            pat = r"^" + pat + r"$"

        self.TEXT = re.sub(pat, new, self.TEXT, flags=flags)
        return self.TEXT

    def sub__regexps(self, *rules: Union[tuple[str], tuple[str, str | None], tuple[str, str | None, re.RegexFlag]], flags: re.RegexFlag = 0, cover_type: PatCoverStyle = PatCoverStyle.NONE) -> str:
        """
        GOAL
        ----

        SPECIALLY CREATED FOR
        ---------------------
        cover_type - for prepare_for_json_parsing
        WORD means syntax word!
        """
        for rule in rules:
            self.sub__regexp(*rule, flags=flags, cover_type=cover_type)

        return self.TEXT

    # -----------------------------------------------------------------------------------------------------------------
    def sub__word(self, *rule, flags: re.RegexFlag = 0) -> str:
        """
        GOAL
        ----
        replace exact word(defined by pattern) in text.
        """
        return self.sub__regexp(*rule, flags=flags, cover_type=PatCoverStyle.WORD)

    def sub__words(self, *rules, flags: re.RegexFlag = 0) -> str:
        """
        GOAL
        ----
        replace exact word(defined by pattern) in text.
        """
        return self.sub__regexps(*rules, flags=flags, cover_type=PatCoverStyle.WORD)

    # -----------------------------------------------------------------------------------------------------------------
    def sub__line(self, *rule, flags: re.RegexFlag = 0) -> str:
        return self.sub__regexp(*rule, flags=flags | re.MULTILINE, cover_type=PatCoverStyle.LINE)

    def sub__lines(self, *rules, flags: re.RegexFlag = 0) -> str:
        return self.sub__regexps(*rules, flags=flags | re.MULTILINE, cover_type=PatCoverStyle.LINE)

    # =================================================================================================================
    def fix__incorrect(self) -> str:
        self.fix__incorrect_quotes()
        self.fix__incorrect_spaces()
        return self.TEXT

    def fix__incorrect_quotes(self) -> str:
        self.TEXT = self.TEXT.replace('”', '"')
        self.TEXT = self.TEXT.replace('“', '"')
        return self.TEXT

    def fix__incorrect_spaces(self) -> str:
        # self.TEXT = self.TEXT.replace('“', ' ')
        return self.TEXT

    def fix__json(self) -> str:
        # PREPARE -----------------------------------------------------------------
        # replace pytonic values (usually created by str(Any)) before attempting to apply json.loads to get original python aux_types
        # so it just same process as re.sub by one func for several values

        # JSON-cmts my special ------------
        self.delete__cmts_c()

        # JSON-COMMAS ------------
        self.clear__regexps(r',\s*(?=\})')

        # JSON-QUOTES ------------
        self.sub__regexp("\'", "\"")

        # JSON-BOOL ------------
        self.sub__word(r"True", "true")
        self.sub__word(r"False", "false")
        self.sub__word(r"None", "null")

        # JSON-dict VALUES ------------
        self.sub__word(r"\s*:\s*\"null\"", ":null")
        self.sub__word(r"\s*:\s*\"true\"", ":true")
        self.sub__word(r"\s*:\s*\"false\"", ":false")

        # NUM KEYS ------------
        self.sub__regexp(r"\b(\d+\.?\d*)\b\s*:\s*", r'"\1":')

        return self.TEXT

    # EDIT ============================================================================================================
    def clear__regexps(self, *pats: str, **kwargs) -> str:
        for pat in pats:
            self.sub__regexp(pat=pat, new="", **kwargs)
        return self.TEXT

    def clear__noneprintable(self) -> str:
        return self.clear__regexps(f"[^{string.printable}а-яА-ЯёЁ]")

    def clear__punctuation(self) -> str:
        return self.clear__regexps(f"[^{string.punctuation}]")

    def clear__spaces_all(self) -> str:
        """
        GOAL
        ----
        make a shortest string for like a str() from any container!
        assert str([1,2]) == "[1, 2]"
        assert func(str([1,2])) == "[1,2]"
        """
        return self.sub__regexp(r" ", "")

    def clear__space_duplicates(self) -> str:
        """
        GOAL
        ----
        replace repetitive spaces by single one
        """
        return self.sub__regexps((r" {2,}", " "))

    def clear__lines(self, *pats: str) -> str:
        """
        NOTE
        ----
        clear! NOT DELETE!!! exact lines!
        if need - apply delete!
        """
        for pat in pats:
            self.sub__line(pat, "")
        return self.TEXT

    def delete__lines_blank(self) -> str:
        """
        GOAL
        ----
        exact deleting blank lines!
        """
        # return self.clear__lines(r"\s*", )

        # variant1
        # self.sub__regexp(r"^\s*\n+", "", re.MULTILINE)        # not enough!
        # self.sub__regexp(r"^\s*\n+", "", re.MULTILINE)        # not enough!

        # variant2
        self.sub__regexp(r"^\s*$", "", re.MULTILINE)        # not enough!
        self.sub__regexp(r"^\s*\n+", "", re.MULTILINE)      # startwith
        self.sub__regexp(r"\n+\s*$", "", re.MULTILINE)      # endswith
        self.sub__regexp(r"\n+\s*\n+", "\n", re.MULTILINE)  # middle double
        return self.TEXT

    # =================================================================================================================
    def delete__cmts(self, cmt_type: CmtStyle = CmtStyle.SHARP) -> str:
        """
        GOAL
        ----
        exact DELETING cmts

        NOTE
        ----
        if one line cmt - full line would be deleted!
        """
        # recursion -----------------------------
        if cmt_type == CmtStyle.ALL:
            for cmt_type in [CmtStyle.SHARP, CmtStyle.DSLASH, CmtStyle.REM, CmtStyle.C]:
                self.delete__cmts(cmt_type)

        elif cmt_type == CmtStyle.AUTO:
            raise NotImplementedError(CmtStyle.AUTO)

        # work ----------------------------------
        if cmt_type == CmtStyle.SHARP:
            self.clear__regexps(PatCmts.SHARP_LINE, flags=re.MULTILINE)
            self.clear__regexps(PatCmts.SHARP_INLINE, flags=re.MULTILINE)

        elif cmt_type == CmtStyle.DSLASH:
            self.clear__regexps(PatCmts.DSLASH_LINE, flags=re.MULTILINE)
            self.clear__regexps(PatCmts.DSLASH_INLINE, flags=re.MULTILINE)

        elif cmt_type == CmtStyle.REM:
            self.clear__regexps(PatCmts.REM_LINE, flags=re.MULTILINE | re.IGNORECASE)    # dont use \s* after REM!!!
            self.clear__regexps(PatCmts.REM_INLINE, flags=re.MULTILINE | re.IGNORECASE)

        elif cmt_type == CmtStyle.C:
            self.clear__regexps(PatCmts.C_MLINE)

        return self.TEXT

    def delete__cmts_sharp(self) -> str:
        return self.delete__cmts(CmtStyle.SHARP)

    def delete__cmts_dslash(self) -> str:
        return self.delete__cmts(CmtStyle.DSLASH)

    def delete__cmts_rem(self) -> str:
        return self.delete__cmts(CmtStyle.REM)

    def delete__cmts_c(self) -> str:
        return self.delete__cmts(CmtStyle.C)

    # =================================================================================================================
    def strip__lines(self) -> str:
        self.lstrip__lines()
        self.rstrip__lines()
        return self.TEXT

    def rstrip__lines(self) -> str:
        """
        GOAL
        ----
        keep indents! strip right!
            " line1 \n line2 " --> " line1\n line2"

        NOTE
        ----
        it can strip blank lines!
            " line1 \n \n  line2 " --> " line1\nline2"
        """
        return self.sub__regexp(r"\s+$", "", re.MULTILINE)

    def lstrip__lines(self) -> str:
        """
        NOTE
        ----
        less usefull as lstrip__lines
        but for the company)
        """
        return self.sub__regexp(r"^\s+", "", re.MULTILINE)

    # =================================================================================================================
    def split_lines(self, skip_blanks: bool = None) -> list[str]:
        lines_all = self.TEXT.splitlines()
        if skip_blanks:
            result_no_blanks = []
            for line in lines_all:
                if line:
                    result_no_blanks.append(line)
            return result_no_blanks

        else:
            return lines_all

    # =================================================================================================================
    def shortcut(
            self,
            maxlen: int = 15,
            where: Where3 = Where3.LAST,
            sub: str | None = "...",
    ) -> str:
        """
        MAIN IDEA-1=for SUB
        -------------------
        if sub is exists in result - means it was SHORTED!
        if not exists - was not shorted!
        """
        sub = sub or ""
        sub_len = len(sub)

        source = self.TEXT
        source_len = len(source)

        if source_len > maxlen:
            if maxlen <= sub_len:
                return sub[0:maxlen]

            if where == Where3.FIRST:
                result = sub + source[-(maxlen - sub_len):]
            elif where == Where3.LAST:
                result = source[0:maxlen - sub_len] + sub
            elif where == Where3.MIDDLE:
                len_start = maxlen // 2 - sub_len // 2
                len_finish = maxlen - len_start - sub_len
                result = source[0:len_start] + sub + source[-len_finish:]
            else:
                result = source
            return result

        return source

    def shortcut_nosub(
            self,
            maxlen: int = 15,
            where: Where3 = Where3.LAST,
    ) -> str:
        """
        GOAL
        ----
        derivative-link for shortcut but no using subs!
        so it same as common slice
        """
        return self.shortcut(maxlen=maxlen, where=where, sub=None)

    # =================================================================================================================
    def findall(self, *pats: str, flags: int = 0) -> list[str]:
        """
        GOAL
        ----
        find all pattern values in text

        NOTE
        ----
        if pattern have group - return group value (as usual)
        """
        result = []
        for pat in pats:
            result_i = re.findall(pat, self.TEXT, flags=flags)
            for value in result_i:
                value: str
                if value == "":
                    continue
                value = value.strip()
                result.append(value)
        return result

    def search__group(self, *pats: str, flags: int = 0) -> str | None:
        """
        GOAL
        ----
        get first found group 1
        if pat without group - just return found value
        None - not found!
        """
        for pat in pats:
            match = re.search(pat, self.TEXT, flags=flags)
            if match:
                try:
                    return match[1]     # group defined in pat
                except:
                    return match[0]     # group NOT defined in pat

        return

    # =================================================================================================================
    def parse__number_single(self, fpoint: TYPE__FPOINT_DRAFT = FPoint.AUTO, num_type: NumType = NumType.BOTH) -> int | float | None:
        """
        GOAL
        ----
        parce single float value (unit available) from text.

        SPECIALLY CREATED FOR
        ---------------------
        UART terminal data validation

        :returns:
            noraise in any case!
            None - no value
            None - value is not single
            None - value is not exact type
        """
        result = None
        if fpoint is not NoValue:
            fpoint = FPoint(fpoint)
        num_type = NumType(num_type)

        # get PAT ---------
        if num_type == NumType.INT:
            pat = PatNumberSingle(fpoint).INT_COVERED
        elif num_type == NumType.FLOAT:
            pat = PatNumberSingle(fpoint).FLOAT_COVERED
        elif num_type == NumType.BOTH:
            pat = PatNumberSingle(fpoint).BOTH_COVERED
        else:
            raise TypeError(f"{num_type=}")

        # FIND STR --------
        match = re.fullmatch(pat, self.TEXT)
        value: str | None = match and match[1]

        # get num ---------
        if value:
            value: str = value.replace(",", ".")

            if num_type == NumType.INT:
                result = int(value)
            elif num_type == NumType.FLOAT:
                result = float(value)
            elif num_type == NumType.BOTH:
                if "." in value:
                    result = float(value)
                else:
                    result = int(value)
        # FINISH ----------
        return result

    def parse__int_single(self) -> int | None:
        return self.parse__number_single(num_type=NumType.INT)

    def parse__float_single(self, fpoint: TYPE__FPOINT_DRAFT = FPoint.AUTO) -> float | None:
        return self.parse__number_single(fpoint=fpoint, num_type=NumType.FLOAT)

    # =================================================================================================================
    def parse__requirements_lines(self) -> list[str]:
        """
        GOAL
        ----
        get list of required modules (actually full lines stripped and commentsCleared)

        SPECIALLY CREATED FOR
        ---------------------
        setup.py install_requires
        """
        self.delete__cmts(CmtStyle.SHARP)
        self.delete__lines_blank()
        self.strip__lines()
        result = self.split_lines()
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def parse__json(self) -> TYPING.DICT_STR_ELEM | TYPING.ELEMENTARY | NoValue:  # NoValue ???? yes to separate from None as parsed object
        """
        NOTE
        ----
        intended source is json dumped! or stringed!

        GOAL
        ----
        create an elementary object from text.
        or return source - FIXME: decide to use - think NO!!!

        by now it works correct only with single elementary values like INT/FLOAT/BOOL/NONE
        for collections it may work but may not work correctly!!! so use it by your own risk and conscious choice!!
        """
        self.fix__json()
        try:
            result = json.loads(self.TEXT)
            return result
        except Exception as exx:
            print(f"{exx!r}")
            return NoValue

    def parse__object_stringed(self) -> TYPING.JSON_ANY | NoValue:
        return self.parse__json()

    # =================================================================================================================
    def parse__dict(self, style: DictTextFormat = DictTextFormat.AUTO) -> TYPING.DICT_STR_ELEM | None:
        if style == DictTextFormat.AUTO:
            return self.parse__dict_auto()

        elif style == DictTextFormat.JSON:
            return self.parse__dict_json()

        elif style == DictTextFormat.INI:
            return self.parse__dict_ini()

        elif style == DictTextFormat.CSV:
            return self.parse__dict_csv()

        else:
            raise NotImplementedError(f"{style=}")

    # -----------------------------------------------------------------------------------------------------------------
    def parse__dict_auto(self) -> TYPING.DICT_STR_ELEM | TYPING.DICT_STR_STR | None:
        for style in [DictTextFormat.JSON, DictTextFormat.INI, DictTextFormat.CSV]:     # order is important!
            result = self.parse__dict(style)
            if result:
                return result

    # -----------------------------------------------------------------------------------------------------------------
    def parse__dict_json(self) -> TYPING.DICT_STR_ELEM | None:     # NoValue ????
        """
        SAME AS parse__json BUT
        -----------------------
        if result is not dict - return None!
        """
        result = self.parse__json()
        if isinstance(result, dict):
            return result

    def parse__dict_ini(self) -> TYPING.DICT_STR_STR | None:
        ini = ConfigParserMod()

        try:
            ini.read_string(self.TEXT)
            return ini.to_dict()
        except Exception as exx:
            msg = f"[CRITICAL] incorrect file!{exx!r}"
            print(msg)
            return

    def parse__dict_csv(self) -> TYPING.DICT_STR_STR | None:
        return
        # raise NotImplemented


# =====================================================================================================================



from typing import *


# =====================================================================================================================
class NestCall_Resolve:
    """
    GOAL
    ----
    just show that nested class used basically for main purpose which will returned by .resolve() method
    its like an aux-function but with better handling

    NOTE
    ----
    dont use it as type!
    dont keep in attributes!
    resolve ti inline!

    SPECIALLY CREATED FOR
    ---------------------
    files.filepath
    """

    def __call__(self, *args, **kwargs) -> Any | NoReturn:
        return self.resolve()

    def resolve(self) -> Any | NoReturn:
        return NotImplemented


# =====================================================================================================================



from typing import *

from base_aux.path1_dir.m2_dir import *
from base_aux.aux_text.m1_text_aux import *


# =====================================================================================================================
class FileAux:
    """
    GOAL
    ----
    single file INTERNAL work!

    textWork use by yourself with TextAux
    """
    FILEPATH: TYPING.PATH_FINAL = None
    TEXT: str = ""       # keep here just for TextAux work!

    def __init__(self, filepath: TYPING.PATH_DRAFT, *args, **kwargs) -> None | NoReturn:
        self.FILEPATH = pathlib.Path(filepath)
        if self.check_exists():
            if self.FILEPATH.is_file():
                # self.read__text()     # NOTE: dont read here! it maybe Bytes! read only in TextFile!
                pass
            else:
                raise Exx__Incompatible(f"{self.FILEPATH=}")
        super().__init__(*args, **kwargs)

    # -----------------------------------------------------------------------------------------------------------------
    def check_exists(self) -> bool:
        return self.FILEPATH and self.FILEPATH.exists()

    def ensure_dir(self) -> None:
        DirAux(self.FILEPATH.parent).create_dirtree()

    def delete_file(self) -> bool:
        self.TEXT = ""
        return DirAux.delete_items(self.FILEPATH)

    def clear_file(self) -> bool:
        self.TEXT = ""
        return self.write__text("") == 0

    # READ/WRITE ======================================================================================================
    # READ ---------------------------------
    def read__text(self) -> Optional[str] | NoReturn:
        if self.FILEPATH.exists():
            if self.FILEPATH.is_file():
                self.TEXT = self.FILEPATH.read_text(encoding="utf-8")
                return self.TEXT
            else:
                raise Exx__Incompatible(f"{self.FILEPATH=}")

    def read__bytes(self) -> Optional[bytes]:
        if self.FILEPATH.exists() and self.FILEPATH.is_file():
            return self.FILEPATH.read_bytes()

    # WRITE ---------------------------------
    def write__text(self, _text: TYPING.STR_DRAFT = None) -> int:
        if _text is not None:
            self.TEXT = str(_text)
        self.ensure_dir()
        return self.FILEPATH.write_text(data=self.TEXT or "", encoding="utf-8")

    def append__text(self, text: TYPING.STR_DRAFT, new_line: bool = False) -> int | NoReturn:
        data = str(text)

        if data:
            self.ensure_dir()

            if self.TEXT and new_line:
                data = f"\n{data}"

            with open(file=self.FILEPATH, encoding="UTF-8", mode="a") as file:  # use exact append for large files!!!
                # if file NOT EXISTS - it creates!!!
                result = file.write(data)
                if result:
                    self.TEXT += data
                return result
        else:
            return True

    def append__lines(self, *lines: TYPING.STR_DRAFT) -> bool | NoReturn:
        data = "\n".join(map(str, lines))
        return bool(self.append__text(data, new_line=True))

    def write__bytes(self, data: bytes) -> Optional[int]:
        if self.FILEPATH:
            self.ensure_dir()
            return self.FILEPATH.write_bytes(data=data)


# =====================================================================================================================



from base_aux.aux_dict.m2_ignorecase import DictIgnorecase
from base_aux.aux_types.m2_info import ObjectInfo

from base_aux.aux_attr.m2_annot2_nest_required import *


# =====================================================================================================================
class DictAttr(DictIgnorecase):
    """
    dot.notation access to dictionary keys.
    RULES
    ------
    1. Levels
        first level - WriteRead
        all other nested levels - ReadOnly for root level! but no errors when set new!
    2. caseInsensitive
    3. when object created it create new copy dict - so it not the same as dict link!




    # FIXME: below is old!
    CANT CREATE SINGLETON!
    cant create singleton!!! i was trying!!! even by typical def __***

    typical usage:
        my_dict = {'param1': 'GOOD'}
        nested_dict = {'param1': 'nested GOOD too'}
        print(my_dict)  # {'param1': 'GOOD'}

        # EXECUTE
        my_dict = DictDotAttrAccess(my_dict)

        # READING
        print(my_dict)  # {'param1': 'GOOD'}
        print(my_dict.param1)  # 'GOOD'
        print(my_dict["param1"])  # 'GOOD'

        # CHANGE
        my_dict.param1 = 123
        print(my_dict)  # {'param1': 123}
        print(my_dict.param1)  # 123
        my_dict["param1"] = 1000
        print(my_dict)  # {'param1': 1000}
        print(my_dict.param1)  # 1000

        # NESTED
        my_dict.nested = nested_dict
        print(my_dict)  # {'param1': 1000, 'nested': {'param1': 'nested GOOD too'}}
        # print(my_dict.nested.param1)  # AttributeError: 'dict' object has no attribute 'param1'

        my_dict.nested = DictDotAttrAccess(nested_dict)
        print(my_dict)  # {'param1': 1000, 'nested': {'param1': 'nested GOOD too'}}
        print(my_dict.nested.param1)  # 'nested GOOD too'

    ОСОБЕННОСТИ:
        +++И ТАК РАБОТАЛО
            1. +BLANK
                my_dict = DictDotAttrAccess()
                print(my_dict)  # {}

            2. +BOOL
                my_dict = DictDotAttrAccess()
                print(bool(my_dict))  # False

                my_dict = DictDotAttrAccess({'p1': '111'})
                print(my_dict)  # {'p1': '111'}
                print(bool(my_dict))  # True

            3. +IN LIST
                my_dict = {'p1': '111'}
                my_dict = DictDotAttrAccess(my_dict)
                print(my_dict)  # {'p1': '111'}

                print("p1" in my_dict)  # True

            5. +при обращении к НЕCУЩЕСТВУЮЩЕМУ - выдает None! без добавления параметра, а после присвоения - параметр создается
                my_dict = {'p1': '111'}
                my_dict = DictDotAttrAccess(my_dict)
                print(my_dict)  # {'p1': '111'}

                print(my_dict.p2)  #None
                print(my_dict)  #{'p1': '111'}
                my_dict.p2 = 222
                print(my_dict)  #{'p1': '111', 'p2': 222}

            6. +UPDATE
                my_dict= DictDotAttrAccess()
                print(my_dict)  # {}

                my_dict.update({1:1})
                print(my_dict)  # {1: 1}

            6. +POP
                my_dict = DictDotAttrAccess({1:1})

                my_dict.pop(1)
                print(my_dict)  # {}

            8. +ITEMS()
                my_dict = DictDotAttrAccess({'p1': '111'})
                print(my_dict.items())  #dict_items([('p1', '111')])

            10. +ИТЕРИРОВАТЬ
                my_dict = {'p1': '111'}
                my_dict = DictDotAttrAccess(my_dict)
                print(my_dict)  # {'p1': '111'}

                for i in my_dict:
                    print(i, my_dict.get(i))    # p1 111

            11. +КОПИЮ
                my_dict = {'p1': '111'}
                my_dict = DictDotAttrAccess(my_dict)
                print(my_dict)  # {'p1': '111'}

                my_dict2 = copy.copy(my_dict)
                print(my_dict2)  # {'p1': '111'}


        ---НЕ ПОЛУЧИЛОСЬ РЕАЛИЗОВАТЬ
            20. замену для __MISSING__
                def __missing__(self, key):
                    return "HELLO"

                __missing__ = lambda: "HELLO"
                __missing__ = FUNC_LINK_LAMBDA_REPEATER
    """
    # -----------------------------------------------------------------------------------------------------------------
    def __getitem__(self, item: Any) -> Any | NoReturn:
        result = self.get(item)         # thats wrong with NESTING! not will working with saving results!!!
        if isinstance(result, dict):
            result = DictAttr(result)

        return result

    # -----------------------------------------------------------------------------------------------------------------
    def __getattr__(self, item: str) -> Any | None:
        return self[item]

    def __setattr__(self, item: str, value: Any) -> None | NoReturn:
        self[item] = value

    def __delattr__(self, item: str | Any) -> None:
        del self[item]


# =====================================================================================================================
class DictAttrAnnotRequired(DictAttr, NestInit_AnnotsRequired):
    """
    its a derivative for DictAttr with applying NestInit_AnnotsRequired

    WHY NOT 1=just simple nesting NestInit_AnnotsRequired?
    --------------------------------------------
    in this case
    first we need apply DICT inition
    and only secondary we could check any by NestInit_AnnotsRequired.
    BUT DICT is a stopping/last class!!!

    so we cant just call init with super()
    """

    def __init__(self, *args, **kwargs) -> None | NoReturn:
        super().__init__(*args, **kwargs)
        # super(NestInit_AnnotsRequired, self).__init__()
        # annot_types = self.annot__get_nested__dict_types()
        # print(f"{annot_types=}")
        self.check_all_defined_or_raise()
        # FIXME: dont use base from NestInit_AnnotsRequired!!! will not work! or fix!

    def check_all_defined_or_raise(self) -> None | NoReturn:
        not_def_list = []
        nested = AnnotsAux(self).get__dict_types()
        for key in nested:
            if key not in self:
                not_def_list.append(key)

        if not_def_list:
            msg = f"{not_def_list=}"
            raise Exception(msg)


# =====================================================================================================================
if __name__ == '__main__':
    class Cls(DictAttrAnnotRequired):
        ATTR1: str

    victim = Cls()
    ObjectInfo(victim).print()


# =====================================================================================================================



from typing import *

from base_aux.lambdas.m1_lambdas import Lambda
from base_aux.base_resolver.m1_resolver import *
from base_aux.aux_attr.m1_attr1_aux import *


# =====================================================================================================================
class NestInit_AttrsLambdaResolve:
    """
    find and call all Lambda aux_attr On class inition
    GOAL
    ----
    if you need create object in classAttribute only on real inition of class
    useful in case of raising exx on init, but you want to pass instance in class attribute with inplace initiation

    REASON EXAMPLE
    --------------
    all class attributes will be calculated on import!
    class Cls:
        OK: int = int("1")
        # FAIL_ON_IMPORT: int = int("hello")    # ValueError: invalid literal for int() with base 10: 'hello'
        FAIL_ON_INIT: int = None

        def __init__(self, *args, **kwargs):
            if self.FAIL_ON_INIT is None:
                self.FAIL_ON_INIT = int("hello")    # this wount raise on import!

    Cls()   # ValueError: invalid literal for int() with base 10: 'hello'
    """

    def __init__(self, *args, **kwargs) -> None | NoReturn:
        for name in AttrAux(self).iter__not_private():
            value = getattr(self, name)
            if isinstance(value, (Lambda, NestCall_Resolve)):
                setattr(self, name, value())

        super().__init__(*args, **kwargs)


# =====================================================================================================================



from typing import *

from base_aux.aux_attr.m1_attr1_aux import *
from base_aux.aux_cmp_eq.m2_eq_aux import *
from base_aux.base_statics.m4_enums import *


# =====================================================================================================================
class NestEq_AttrsNotPrivate:
    """
    GOAL
    ----
    mainly used for cmp bare attr-kits with no callables!

    LOGIC
    -----
    cmp first - direct callables
    cmp second - resolveExx!!!
    """
    def __eq__(self, other: Any) -> bool:
        # if isinstance() NestInit_AnnotsAttrByKwArgs == NestInit_AnnotsAttrByKwArgs:
        #     # check by names

        if other is None:
            return False

        try:
            for attr in AttrAux(self).iter__not_private():
                # 1=cmp direct --------
                value_self_direct = AttrAux(self).getattr__callable_resolve(attr, CallableResolve.DIRECT)
                value_other_direct = AttrAux(other).getattr__callable_resolve(attr, CallableResolve.DIRECT)
                if EqAux(value_self_direct).check_doubleside__bool(value_other_direct):
                    continue

                # 2=cmp callables --------      # TODO: use CallableResolve.SKIPCALLABLES ???
                value_self = AttrAux(self).getattr__callable_resolve(attr, CallableResolve.EXX)
                value_other = AttrAux(other).getattr__callable_resolve(attr, CallableResolve.EXX)

                if not EqAux(value_self).check_doubleside__bool(value_other):
                    return False

            return True
        except:
            return False


# =====================================================================================================================



import pytest

from base_aux.alerts.m1_alerts1_smtp import *
from base_aux.alerts.m1_alerts2_telegram import *
from base_aux.aux_callable.m1_callable_aux import *


# =====================================================================================================================
# TODO: separate by base class and use victimCls as attr!

@pytest.mark.skipif(CallableAux(AlertSmtp.AUTH).check_raise(), reason="no file")
class Test__1:
    @pytest.mark.parametrize(argnames="victim", argvalues=[
        AlertSmtp,
        AlertTelegram,
    ])
    def test__send_single(self, victim):
        assert victim("single").result_wait() is True

    @pytest.mark.parametrize(argnames="victim", argvalues=[AlertSmtp, AlertTelegram])
    def test__send_multy_thread(self, victim):
        threads = [
            victim("thread1"),
            victim("thread2"),
            victim("thread3"),
        ]

        victim.threads_wait_all()

        for thread in threads:
            assert thread._result is True

    @pytest.mark.parametrize(argnames="victim", argvalues=[AlertSmtp, AlertTelegram])
    @pytest.mark.parametrize(argnames="_subj_name, body, _subtype", argvalues=[
        (None, "zero", None),
        ("", "plain123", "plain123"),
        ("plain", "plain", "plain"),
        ("html", "<p><font color='red'>html(red)</font></p>", "html")
    ])
    def test__send_single__parametrized(self, victim, _subj_name, body, _subtype):
        assert victim(_subj_name=_subj_name, body=body, _subtype=_subtype).result_wait() is True

    @pytest.mark.parametrize(argnames="victim", argvalues=[AlertSmtp, AlertTelegram])
    def test__send_multy__result_wait(self, victim):
        assert victim("multy1").result_wait() is True
        assert victim("multy2").result_wait() is True

    @pytest.mark.parametrize(argnames="victim", argvalues=[AlertSmtp, AlertTelegram])
    def test__send_multy__wait_join(self, victim):
        thread1 = victim("thread1")
        thread2 = victim("thread2")

        thread1.wait()
        thread2.wait()

        assert thread1._result is True
        assert thread2._result is True


# =====================================================================================================================
# class Test_AlertHttp:
#     Victim = AlertHttp
#     def test__send_single(self, victim):
#         assert self.Victim({}).result_wait() is True


# =====================================================================================================================



from base_aux.classes.m2_middle_group import *

from .tc_types import TYPE__RESULT_BASE, TYPE__RESULT_W_NORETURN, TYPE__RESULT_W_EXX


# =====================================================================================================================
class TcGroup_Base(ClsMiddleGroup):
    """
    make groups by separating startup/teardown
    """
    MIDDLE_GROUP__NAME = None
    MIDDLE_GROUP__CMP_ATTR = None

    @classmethod
    def startup__cls__wrapped(cls) -> TYPE__RESULT_W_NORETURN:
        return True

    @classmethod
    def teardown__cls__wrapped(cls) -> TYPE__RESULT_W_NORETURN:
        return True

    # @classmethod
    # def middle_group__check_equal__cls(cls, other: type[Self]) -> bool:
    #     """
    #
    #     this is will not work on classmethods!!!
    #     -------------------------
    #         class Cls:
    #             def meth(self):
    #                 pass
    #             @classmethod
    #             def cmeth(cls):
    #                 pass
    #
    #         class Cls2(Cls):
    #             pass
    #
    #         print(Cls.meth is Cls2.meth)
    #         print(Cls.meth == Cls2.meth)
    #         print(Cls.cmeth is Cls2.cmeth)
    #         print(Cls.cmeth == Cls2.cmeth)
    #
    #         print(Cls.meth)
    #         print(Cls2.meth)
    #         print(Cls.cmeth)
    #         print(Cls2.cmeth)
    #
    #         #
    #         True
    #         True
    #         False
    #         False
    #         <function Cls.meth at 0x000002CC6B525940>
    #         <function Cls.meth at 0x000002CC6B525940>
    #         <bound method Cls.cmeth of <class '__main__.Cls'>>
    #         <bound method Cls.cmeth of <class '__main__.Cls2'>>
    #     -------------------------
    #     """
    #     try:
    #         return cls.startup__cls__wrapped is other.startup__cls__wrapped and cls.teardown__cls__wrapped is other.teardown__cls__wrapped
    #     except:
    #         pass
    #     return False


# =====================================================================================================================



from typing import *
from enum import Enum, auto
from dataclasses import dataclass

from base_aux.aux_attr.m2_annot1_aux import *


# =====================================================================================================================
class IndicatorName(Enum):
    WMA = auto()
    STOCH = auto()
    ADX = auto()
    MACD = auto()
    RSI = auto()


# ---------------------------------------------------------------------------------------------------------------------
class IndicatorParamsBase:
    NAME: IndicatorName = None
    COLUMN_NAME__TEMPLATE: str = None
    ROUND: int = None

    def __iter__(self):
        yield from self.params_dict__get().values()

    def column_name__get(self) -> str:
        return self.COLUMN_NAME__TEMPLATE % self.params_dict__get()

    def params_dict__get(self):
        return AnnotsAux(self).get__dict_values()

    def bars_expected__get(self) -> int:
        """

        РАСЧЕТ ДЛИНЫ БАРОВ
            количество влияет на результат!!!!
            при не особо достаточном количестве баров - расчет произойдет НО значения будут отличаться от фактического!!!
            видимо из-за того что будет вычитаться с нулевыми некоторыми начальными значениями!!!

            sum * 2 = это очень мало!!!!!
            sum * 10 = кажется первая, что вообще показала полное совпадение с Tinkoff терминалом!!!

            ADX
                !!! ЭТО ОЧЕНЬ ВАЖНО ДЛЯ ADX !!!!
            STOCH
                вообще не важно - кажется там сколько длина его - столько и баров достаточно!!!
        """
        return sum(self) * 10


# ---------------------------------------------------------------------------------------------------------------------
@dataclass
class IndicatorParams_WMA(IndicatorParamsBase):
    length: int

    NAME: IndicatorName = IndicatorName.WMA
    # FUNCTION: Callable = ta.wma     # NOT WORKING!
    COLUMN_NAME__TEMPLATE: str = "WMA_%(length)s"
    ROUND: int = 1


@dataclass
class IndicatorParams_STOCH(IndicatorParamsBase):
    """
    always work with 14/3/3!!!
    """
    fast_k: int
    slow_k: int
    slow_d: int

    NAME: IndicatorName = IndicatorName.STOCH
    COLUMN_NAME__TEMPLATE: str = "STOCHk_%(fast_k)s_%(slow_k)s_%(slow_d)s"
    # COLUMN_NAME__TEMPLATE: str = "STOCHk_14_3_3"
    ROUND: int = 1


@dataclass
class IndicatorParams_ADX(IndicatorParamsBase):
    length: int
    lensig: int

    NAME: IndicatorName = IndicatorName.ADX

    COLUMN_NAME__TEMPLATE: str = "ADX_%(lensig)s"
    ROUND: int = 1


@dataclass
class IndicatorParams_MACD(IndicatorParamsBase):
    fast: int
    slow: int
    signal: int

    NAME: IndicatorName = IndicatorName.MACD
    ROUND: int = 3

    @property
    def COLUMN_NAME__TEMPLATE(self) -> str:
        if self.slow < self.fast:
            return "MACDh_%(slow)s_%(fast)s_%(signal)s"
        else:
            return "MACDh_%(fast)s_%(slow)s_%(signal)s"


@dataclass
class IndicatorParams_RSI(IndicatorParamsBase):
    length: int

    NAME: IndicatorName = IndicatorName.RSI

    COLUMN_NAME__TEMPLATE: str = "RSI_%(length)s"
    ROUND: int = 1


# =====================================================================================================================



from base_aux.aux_cmp_eq.m1_cmp import *
from base_aux.aux_cmp_eq.m10_types import *
from base_aux.aux_text.m1_text_aux import *


# =====================================================================================================================
TYPE__VERSION_BLOCK_ELEMENT = Union[str, int]
TYPE__VERSION_BLOCK_ELEMENTS_FINAL = tuple[TYPE__VERSION_BLOCK_ELEMENT, ...]
TYPE__VERSION_BLOCK_ELEMENTS_DRAFT = Union[str, int, list[TYPE__VERSION_BLOCK_ELEMENT],  TYPE__VERSION_BLOCK_ELEMENTS_FINAL, Any, 'VersionBlock']


# =====================================================================================================================
class VersionBlock(NestCmp):
    """
    this is exact block in version string separated by dots!!!

    PATTERN for blocks
    ------------------
        block1.block2.block3

    EXAMPLES for block
    ------------------
        1rc2
        1-rc2
        1 rc 2

    RULES
    -----
    1.
    """
    SOURCE: TYPE__VERSION_BLOCK_ELEMENTS_DRAFT
    ELEMENTS: TYPE__VERSION_BLOCK_ELEMENTS_FINAL = ()
    RAISE: bool = True
    EQ_VALID: TYPE__EQ_VALID = EqValidChain(
        # EqValid_RegexpAnyTrue(*PatVersionBlock.VALID),
        EqValid_RegexpAllFalse(*PatVersionBlock.VALID_REVERSE),
    )

    def __init__(
            self,
            source: TYPE__VERSION_BLOCK_ELEMENTS_DRAFT,
            eq_valid: TYPE__EQ_VALID = None,
            _raise: bool = None,
    ) -> None | NoReturn:
        if eq_valid is not None:
            self.EQ_VALID = eq_valid

        if _raise is not None:
            self.RAISE = _raise

        self.SOURCE = source
        if self.SOURCE is None:
            self.SOURCE = ""

        self._prepare_source()
        self._parse_elements()

    def _prepare_source(self) -> str:
        if isinstance(self.SOURCE, (list, tuple)):
            result = "".join([str(item) for item in self.SOURCE])
        else:
            result = str(self.SOURCE)

        # FINISH -------------------------------
        result = result.lower()
        result = result.strip()

        self.SOURCE = result
        return result

    def _parse_elements(self) -> TYPE__VERSION_BLOCK_ELEMENTS_FINAL | NoReturn:
        if self.EQ_VALID and self.EQ_VALID != self.SOURCE:
            if self.RAISE:
                raise Exx__Incompatible(f"{self.SOURCE=}/{self.EQ_VALID=}")
            else:
                return ()

        result_list = []
        for element in re.findall(PatVersionBlock.ITERATE, self.SOURCE):
            try:
                element = int(element)
            except:
                pass
            result_list.append(element)

            if len(result_list) > 1:
                if type(result_list[-1]) == type(result_list[-2]):
                    if self.RAISE:
                        raise Exx__Incompatible(f"{result_list[-1]=}/{result_list[-2]=}")
                    else:
                        result_list = ()
                        break

        result = tuple(result_list)
        self.ELEMENTS = result
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self})"

    def __str__(self) -> str:
        return "".join(str(item) for item in self)

    # -----------------------------------------------------------------------------------------------------------------
    def __iter__(self):
        yield from self.ELEMENTS

    def __len__(self) -> int:
        return len(self.ELEMENTS)

    def __bool__(self) -> bool:
        """
        GOAL
        ----
        exists NOT ZERO velues,
        """
        if len(self) == 0:
            return False

        for elem in self:
            if elem:
                return True

        return False

    # CMP -------------------------------------------------------------------------------------------------------------
    def __cmp__(self, other) -> int | NoReturn:
        other = self.__class__(
            other,
            # eq_valid=self.EQ_VALID,
            _raise=False
        )

        # equel ----------------------
        if not self and not other:
            return 0

        if str(self) == str(other):
            return 0

        # by elements ----------------
        for elem_1, elem_2 in zip(self, other):
            if elem_1 == elem_2:
                continue

            if isinstance(elem_1, int):
                if isinstance(elem_2, int):
                    return elem_1 - elem_2
                else:
                    return 1
            else:
                if isinstance(elem_2, int):
                    return -1
                else:
                    return int(elem_1 > elem_2) or -1

        # final - longest ------------
        return int(len(self) > len(other)) or -1


# =====================================================================================================================
TYPE__VERSION_BLOCKS_FINAL = tuple[VersionBlock, ...]


# =====================================================================================================================



from base_aux.base_statics.m2_exceptions import Exx__ValueNotValidated

from base_aux.aux_cmp_eq.m2_eq_aux import *
from base_aux.aux_cmp_eq.m4_eq_valid_chain import *
from base_aux.base_resolver.m1_resolver import *


# =====================================================================================================================
class ValueEqValid(NestCall_Resolve):
    """
    GOAL
    ----
    class to use validation by Eq objects for new values

    NOTE
    ----
    universal - need to pass EQ object!
    """
    __value: Any = NoValue
    VALUE_DEFAULT: Any = NoValue
    EQ: EqValid_Base | type[EqValid_Base] | EqValidChain | type[NoValue] = NoValue

    def __init__(
            self,
            value: Any = NoValue,
            eq: EqValid_Base | type[EqValid_Base] | EqValidChain | type[NoValue] = NoValue,
            eq_args: TYPING.ARGS_DRAFT = ARGS_FINAL__BLANK,          # NOTE: dont try to use INDIRECT style passing */**
            eq_kwargs: TYPING.KWARGS_DRAFT = KWARGS_FINAL__BLANK,
    ) -> None | NoReturn:
        if eq is not NoValue:
            self.EQ = eq

        if TypeAux(self.EQ).check__class() and issubclass(self.EQ, EqValid_Base):
            self.EQ = self.EQ(*eq_args, **eq_kwargs)

        if value is not NoValue:
            self.VALUE = value

        self.VALUE_DEFAULT = self.VALUE

    def __str__(self) -> str:
        return f"{self.VALUE}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.VALUE},eg={self.EQ})"

    def __eq__(self, other) -> bool:
        if isinstance(other, ValueEqValid):
            other = other.VALUE

        return EqAux(self.VALUE).check_doubleside__bool(other)

    def resolve(self) -> Any:
        return self.VALUE

    @property
    def VALUE(self) -> Any:
        return self.__value

    @VALUE.setter
    def VALUE(self, value: Any) -> Optional[NoReturn]:
        if self.EQ == value or self.EQ is NoValue:    # place EQ at first place only)
            self.__value = value
        else:
            raise Exx__ValueNotValidated()

    def reset(self, value: Any | NoValue = NoValue) -> bool | NoReturn:
        """
        set new value or default
        """
        if value == NoValue:
            self.VALUE = self.VALUE_DEFAULT
        else:
            self.VALUE = value

        return True     # True - is for success only!


# =====================================================================================================================
class ValueEqValid_Exact(ValueEqValid):
    """
    GOAL
    ----
    base class to make a classes with specific validation

    SAME AS - ValueEqValid but
    --------------------------
    all args/kwargs passed into EQ

    NOTE
    ----
    exact EQ! - no need to pass EQ object! already in class
    """
    EQ: type[EqValid_Base]

    def __init__(
            self,
            value: Any,
            *eq_args: TYPING.ARGS_DRAFT,
            **eq_kwargs: TYPING.KWARGS_DRAFT,
    ) -> None:
        super().__init__(value=value, eq=NoValue, eq_args=eq_args, eq_kwargs=eq_kwargs)


# ---------------------------------------------------------------------------------------------------------------------
@final
class ValueEqValid_Variants(ValueEqValid_Exact):
    """
    SAME AS - ValueVariants but
    ---------------------------
    here is only validating and keep passed value
    in ValueVariants - final value used from exact Variants!
    """
    EQ = EqValid_VariantsDirect


@final
class ValueEqValid_VariantsStrLow(ValueEqValid_Exact):
    EQ = EqValid_VariantsStrLow


# =====================================================================================================================
if __name__ == "__main__":
    assert ValueEqValid_Variants(1, *(1, 2))
    try:
        assert ValueEqValid_Variants(1, *(10, 2))
        assert False
    except:
        assert True

    try:
        assert ValueEqValid_Variants("val", *("VAL", 2))
        assert False
    except:
        assert True

    assert ValueEqValid_VariantsStrLow(1, *(1, 2))
    assert ValueEqValid_VariantsStrLow("val", *("VAL", 2))


# =====================================================================================================================



from base_aux.stock.m4_strategy import *


class MonitorADX(MonitorADX):
    THRESH_ADX_FAST_FULL = 80
    THRESH_ADX_SLOW_FULL = 50


def main():
    MonitorADX().join()
    # trade_alerts.IndicatorMapDrawer_Simple().join()


if __name__ == "__main__":
    main()
    


from typing import *

from base_aux.aux_attr.m1_attr1_aux import AttrAux


# =====================================================================================================================
class NestStR_Attrs:
    """
    GOAL
    ----
    apply str/repr for show attrs names+values

    CAREFUL
    -------
    dont use in Nest* classes - it can used only in FINALs!!! cause it can have same or meaning is not appropriate!
    """
    def __str__(self):  # it can used only in FINAL!!! dont use in Nest*!!!
        result = ""
        for attr_name in AttrAux(self).iter__not_private():
            if result:
                result += ","
            elif not result:
                result += f"{self.__class__.__name__}("
            value = getattr(self, attr_name)
            result += f"{attr_name}={value}"

        result += ")"
        return result

    def __repr__(self):       # it can used only in FINAL!!! dont use in Nest*!!!
        return str(self)


# =====================================================================================================================



from typing import *

from base_aux.base_inits.m1_nest_init_source import *


# =====================================================================================================================
# TODO: use as ValueExplicit
# FIXME: deprecate???

@final
class Explicit(NestInit_Source):
    """
    GOAL
    ----
    1. solve NONE-VALUE as ambiguity by simple way
    2. show/pass explicitly the EXACT VALUE like None/[]/()

    RULES (when apply as funcResult)
    -----
    return object if you get final result!
    return None if there are any errors withing execution

    OLD ----

    WHY NOT-1: NamedTuple
    ----------------
    cause of we need to be able to compare different aux_types by values.
    maybe we need just add __eq__ method instead of it!!!

    USAGE
    -----
    to get exact value - just CALL instance! dont use .VALUE attr-access by attribute!

        from funcs import *

        def func(a, b) -> Optional[Explicit]:
            if a in b:
                return Explicit(a)
            else:
                return

        result = func("None", [None, ])
        assert result is None

        result = func(None, [None, ])
        assert result == Explicit(None)

        if result:
            print(result())         # None
    """
    SOURCE: Any   # dont use SOURCE access by attribute

    def __call__(self, *args, **kwargs) -> Any:
        return self.SOURCE

    def __str__(self):
        return f"{self.__class__.__name__}({self.SOURCE})"

    # -----------------------------------------------------------------------------------------------------------------
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.SOURCE == other()
        else:
            return self.SOURCE == other


# =====================================================================================================================



from typing import *
# =====================================================================================================================
class PROJECT:
    # AUTHOR -----------------------------------------------
    AUTHOR_NAME: str = "Andrei Starichenko"
    AUTHOR_EMAIL: str = "centroid@mail.ru"
    AUTHOR_HOMEPAGE: str = "https://github.com/centroid457/"

    # PROJECT ----------------------------------------------
    NAME_IMPORT: str = "threading_manager"
    KEYWORDS: List[str] = [
        "threading", "threads", "thread",
        "thread manager",
    ]
    CLASSIFIERS_TOPICS_ADD: List[str] = [
        # "Topic :: Communications",
        # "Topic :: Communications :: Email",
    ]

    # README -----------------------------------------------
    # add DOUBLE SPACE at the end of all lines! for correct representation in MD-viewers
    DESCRIPTION_SHORT: str = "manage spawned threads"
    DESCRIPTION_LONG: str = """
    Designed to working with spawned threads

    NOTE: maybe you dont need use it if you need only one class method - use direct QThread
    """
    FEATURES: List[str] = [
        # "feat1",
        # ["feat2", "block1", "block2"],

        "use different managers for different funcs/methods if needed",
        "use just one decorator to spawn threads from func / methods",
        "keep all spawned threads in list by ThreadItem aux_types",
        "ThreadItem keeps result/exx/is_alive attributes!",
        "use wait_all/terminate_all()",
    ]

    # HISTORY -----------------------------------------------
    VERSION: Tuple[int, int, int] = (0, 0, 9)
    TODO: List[str] = [
        "add SERIAL execution as method wait_all_piped! paired up with wait_all_parallel()",
        "add meta cumulative funks",
        "add GROUP threads - in decorator+wait+...",
        "maybe AUTO CLEAR if decorator get new funcName?",
        "TIME item+group",
    ]
    FIXME: List[str] = [
        "..."
    ]
    NEWS: List[str] = [
        "[__INIT__.py] fix import",
        "apply last pypi template",
    ]

    # FINALIZE -----------------------------------------------
    NAME_INSTALL: str = NAME_IMPORT.replace("_", "-")


# =====================================================================================================================



from base_aux.base_statics.m1_types import *
from base_aux.aux_text.m1_text_aux import *
from base_aux.base_inits.m1_nest_init_source import *


# =====================================================================================================================
# @final
class ValidAux(NestInit_Source):
    """
    Try to keep all validating funcs in separated place
    """
    # FIXME: do smth! before it was all static! and source at first place!
    SOURCE: Any = None

    def eq(self, other: Any) -> bool | NoReturn:
        return NotImplemented

    def ltgt(self, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        result = True
        # low ----------
        if low is not None:
            result &= self.SOURCE > low
        if not result:
            return False
        # high ----------
        if high is not None:
            result &= self.SOURCE < high
        # final ----------
        return result

    def ltge(self, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        result = True
        # low ----------
        if low is not None:
            result &= self.SOURCE > low
        if not result:
            return False
        # high ----------
        if high is not None:
            result &= self.SOURCE <= high
        # final ----------
        return result

    def legt(self, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        result = True
        # low ----------
        if low is not None:
            result &= self.SOURCE >= low
        if not result:
            return False
        # high ----------
        if high is not None:
            result &= self.SOURCE < high
        # final ----------
        return result

    def lege(self, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        result = True
        # low ----------
        if low is not None:
            result &= self.SOURCE >= low
        if not result:
            return False
        # high ----------
        if high is not None:
            result &= self.SOURCE <= high
        # final ----------
        return result


# =====================================================================================================================
@final
class ValidAux_Obj(ValidAux):
    """
    cmp will be executed by direct object without modification
    """
    SOURCE: Any = None


# =====================================================================================================================
@final
class ValidAux_NumParsedSingle(ValidAux):
    """
    cmp will be executed by parse single num from STR(source)
    """
    SOURCE: TYPES.NUMBER | None = None

    def init_post(self) -> None:
        self.SOURCE = TextAux(self.SOURCE).parse__number_single()

    def eq(self, other: Any | None | bool | Union[NumType | int | float]) -> bool:
        if other is True:
            return self.SOURCE is not None

        elif other in [False, None, NoValue]:
            return self.SOURCE is None

        elif other in NumType:
            return TextAux(self.SOURCE).parse__number_single(num_type=other) is not None

        else:
            return self.SOURCE == TextAux(other).parse__number_single()


# =====================================================================================================================



import time
from typing import *
import requests
from collections import deque
from enum import Enum, auto

from PyQt5.QtCore import QThread

from base_aux.loggers.m1_logger import Logger
from .m0_url import UrlCreator


# =====================================================================================================================
TYPE__RESPONSE = Union[None, requests.Response, requests.ConnectTimeout]
TYPE__REQUEST_BODY = Union[str, dict]


class ResponseMethod(Enum):
    POST = auto()
    GET = auto()


# =====================================================================================================================
class Client_RequestItem(Logger, UrlCreator, QThread):
    """
    DONT USE IT AS ONE INSTANCE FOR SEVERAL REQUESTS!!!
    You need keep it only to manage results or sent in further time!

    So Only ONE REQUESTITEM FOR ONE Request!

    create object and wait result by wait() or connect slot finished

    :ivar TIMEOUT_SEND: be careful! it is not so clear! timeout for connection is basically 3 sec!

    """
    # SETTINGS -------------------------------------
    START_ON_INIT: bool = None      # DONT DELETE!!! useful for delayed/pending requests
    TIMEOUT_SEND: float = 1

    RETRY_LIMIT: int | None = 1
    RETRY_TIMEOUT: float = 0.5
    retry_index: int = 0
    SUCCESS_IF_FAIL_CODE: bool | None = None    # need to separate timedout and fail statuscode

    METHOD: ResponseMethod = ResponseMethod.POST

    # INIT ------------------------------------------
    BODY: Optional[TYPE__REQUEST_BODY]
    # REQUEST: Optional[requests.Request]
    RESPONSE: Optional[requests.Response]
    EXX: Union[None, requests.ConnectTimeout, Exception]
    TIMESTAMP: float

    # AUX ------------------------------------------
    INDEX: int | None = None

    def __init__(
            self,
            body: Optional[TYPE__REQUEST_BODY] = None,
            method: Optional[ResponseMethod] = None,

            # url: Optional[str] = None,
            host: Optional[str] = None,
            port: Optional[int] = None,
            route: Optional[str] = None,
    ):
        super().__init__()

        # INITS ------------------------------------
        if method is not None:
            self.METHOD = method

        self.BODY = body
        self.RESPONSE = None
        self.EXX = None
        self.TIMESTAMP = 0

        # if url is None:
        #     url = self.HOST
        if host is not None:
            self.HOST = host
        if port is not None:
            self.PORT = port
        if route is not None:
            self.ROUTE = route

        self.index__init()

        # START ------------------------------------
        if self.START_ON_INIT:
            self.start()

    def index__init(self) -> None:
        if self.__class__.INDEX is None:
            self.__class__.INDEX = 0
        else:
            self.__class__.INDEX += 1

        self.INDEX = self.__class__.INDEX

    def check_success(self) -> bool:
        if self.RESPONSE is None:
            return False

        if self.SUCCESS_IF_FAIL_CODE:
            return True
        else:
            return self.RESPONSE.ok

    def __str__(self) -> str:
        return f"[{self.INDEX=}/len={self.__class__.INDEX+1}/{self.retry_index=}/{self.check_success()=}]{self.EXX=}/{self.RESPONSE=}"

    def __repr__(self) -> str:
        return str(self)

    # ------------------------------------------------------------------------------------------------
    def start(self, *args):
        """
        apply only one thread at once (from stack)!
        """
        if not self.isRunning():
            self.LOGGER.debug("start")

            super().start(*args)

    def run(self) -> None:
        self.LOGGER.debug("run")
        while True:
            self._send()

            # CHECK EXIT ----------------------------------------------
            if self.check_success():
                return

            # retry LIMIT ---------------------
            if self.RETRY_LIMIT and self.retry_index == self.RETRY_LIMIT - 1:
                return

            self.retry_index += 1
            time.sleep(self.RETRY_TIMEOUT)

    def _send(self) -> None:
        self.TIMESTAMP = time.time()
        self.RESPONSE = None
        self.EXX = None

        url = self.URL_create()

        with requests.Session() as session:
            try:
                if self.METHOD == ResponseMethod.POST:
                    self.RESPONSE = session.post(url=url, json=self.BODY or {}, timeout=self.TIMEOUT_SEND)
                elif self.METHOD == ResponseMethod.GET:
                    self.RESPONSE = session.get(url=url, timeout=self.TIMEOUT_SEND)
            except Exception as exx:
                self.EXX = exx

        self.LOGGER.debug(self)


# =====================================================================================================================
class Client_RequestsStack(Logger, QThread):
    # SETTINGS -------------------------------------
    REQUEST_CLS: type[Client_RequestItem] = Client_RequestItem

    # AUX ------------------------------------------
    __stack: deque

    def __init__(self):
        super().__init__()
        self.__stack = deque()

    @property
    def stack(self) -> deque:
        return self.__stack

    @property
    def request_active(self) -> Optional[Client_RequestItem]:
        if self.stack:
            return self.stack[0]

    # ------------------------------------------------------------------------------------------------
    def start(self, *args):
        """
        apply only one thread at once (from stack)!
        """
        if not self.isRunning():
            self.LOGGER.debug(f"start {self.request_active}")
            super().start(*args)

    # ------------------------------------------------------------------------------------------------
    def run(self):
        self.LOGGER.debug("[STACK]run")

        # WORK -----------------------------------------
        while len(self.stack):
            self.LOGGER.debug(f"[STACK]run cycle with len={len(self.stack)}")
            self.request_active.run()

            if self.request_active.check_success():
                self.stack.popleft()
            else:
                break

        # FINISH -----------------------------------------
        if self.check_success():
            self.LOGGER.info(f"[STACK] is empty")
        else:
            self.LOGGER.warn(f"[STACK] is stopped [at len={len(self.stack)}] by some errors [exx={self.request_active.EXX=}]")

    def send(self, **kwargs) -> None:
        """
        work usually with POST
        """
        item = self.REQUEST_CLS(**kwargs)
        self.stack.append(item)
        self.LOGGER.debug(f"[STACK].APPEND len={len(self.stack)}")
        self.start()

    def check_success(self) -> bool:
        result = self.request_active is None
        self.LOGGER.debug(result)
        return result


# =====================================================================================================================



import json
import datetime
from PyQt5.QtCore import QThread, pyqtSignal

from base_aux.aux_dict.m3_dict_attr1_simple import *
from base_aux.aux_callable.m1_callable_aux import *
from base_aux.valid.m2_valid_base import *
from base_aux.pyqt.m0_signals import *
from base_aux.loggers.m1_logger import *

from .tc_types import TYPE__RESULT_BASE, TYPE__RESULT_W_NORETURN, TYPE__RESULT_W_EXX
from .models import *
from .tc_groups import *


# =====================================================================================================================
class _TestCaseBase0(Logger):
    # just to use in Signals before defining exact
    pass


# =====================================================================================================================
class Signals(SignalsTemplate):
    signal__tc_state_changed = pyqtSignal(_TestCaseBase0)


# =====================================================================================================================
class _TestCaseBase(TcGroup_Base, _TestCaseBase0, QThread):
    LOG_ENABLE = False
    LOG_USE_FILE = False

    # SETTINGS ------------------------------------
    NAME: str = ""      # set auto!
    DESCRIPTION: str = ""
    SKIP: Optional[bool] = None     # access only over CLASS attribute! not instance!!!
    skip_tc_dut: Optional[bool] = None
    ASYNC: Optional[bool] = True
    # STOP_IF_FALSE_RESULT: Optional[bool] = None     # NOT USED NOW! MAYBE NOT IMPORTANT!!!
    SETTINGS_FILES: Union[None, pathlib.Path, list[pathlib.Path]] = None

    DEVICES__BREEDER_CLS: type['DevicesBreeder'] = None

    # AUXILIARY -----------------------------------
    finished__cls: bool | None = None

    signals: Signals = Signals()  # FIXME: need signal ON BASE CLASS! need only one SlotConnection! need Singleton?
    _INSTS_DICT_CLS: dict[type[Any], dict[Any, Any]]

    result__startup_cls: TYPE__RESULT_BASE = None
    result__teardown_cls: TYPE__RESULT_BASE = None

    # INSTANCE ------------------------------------
    _inst_inited: Optional[bool] = None

    INDEX: int
    SETTINGS: DictAttr = {}
    DEVICES__BREEDER_INST: 'DevicesBreeder'

    result__startup: TYPE__RESULT_W_EXX = None
    result__teardown: TYPE__RESULT_W_EXX = None

    _result: TYPE__RESULT_W_EXX = None
    _timestamp_last: Optional[datetime.datetime]
    timestamp_start: Optional[datetime.datetime]
    details: dict[str, Any]
    exx: Optional[Exception]
    progress: int

    # =================================================================================================================
    @classmethod
    def devices__apply(cls, devices_cls: type['DevicesBreeder'] = None) -> None:
        if devices_cls is not None:
            cls.DEVICES__BREEDER_CLS = devices_cls
            cls.TCS__LIST.clear()

        if cls.DEVICES__BREEDER_CLS:
            cls.DEVICES__BREEDER_CLS.generate__objects()
            if not cls.TCS__LIST:
                for index in range(cls.DEVICES__BREEDER_CLS.COUNT):
                    tc_inst = cls(index=index)

    @classmethod
    @property
    def TCS__LIST(cls) -> list[Self]:
        try:
            result = list(_TestCaseBase._INSTS_DICT_CLS[cls].values())
        except:
            result = []
        return result

    # =================================================================================================================
    def __new__(cls, index: int, *args, **kwargs):
        """
        use singletons for every class!
        """
        if not hasattr(_TestCaseBase, "_INSTS_DICT_CLS"):
            setattr(_TestCaseBase, "_INSTS_DICT_CLS", {})

        if cls not in _TestCaseBase._INSTS_DICT_CLS:
            _TestCaseBase._INSTS_DICT_CLS.update({cls: {}})

        INSTS_DICT = _TestCaseBase._INSTS_DICT_CLS[cls]

        try:
            INST = INSTS_DICT[index]
        except:
            INST = super().__new__(cls)
            INSTS_DICT[index] = INST

        print(f"{cls.__name__}.__NEW__={index=}/{args=}/{kwargs=}//groups={len(_TestCaseBase._INSTS_DICT_CLS)}/group={len(INSTS_DICT)}")
        return INST

    def __init__(self, index: int):
        if self._inst_inited:
            return

        # NEW INSTANCE -----------------------------
        self.INDEX = index
        self.clear()
        super().__init__()

        if self.DEVICES__BREEDER_CLS:
            self.DEVICES__BREEDER_INST = self.DEVICES__BREEDER_CLS(index)

        self.SETTINGS = DictAttr(self.settings_read())
        self._inst_inited = True

    # =================================================================================================================
    @property
    def timestamp_last(self) -> datetime.datetime | None:
        """
        None - not even started
        float - was started!
            stable - finished
            UnStable - in progress (active thread)
        """
        if self._timestamp_last:
            return self._timestamp_last

        if self.isRunning():
            return datetime.datetime.now()

    @timestamp_last.setter
    def timestamp_last(self, value: datetime.datetime | None) -> None:
        self._timestamp_last = value

    # =================================================================================================================
    @classmethod
    def settings_read(cls, files: Union[None, pathlib.Path, list[pathlib.Path]] = None) -> dict:
        result = {}

        _settings_files = files or cls.SETTINGS_FILES
        if _settings_files:
            if isinstance(_settings_files, pathlib.Path):
                _settings_files = [_settings_files, ]

            if isinstance(_settings_files, (list, tuple)):
                for file in _settings_files:
                    if file.exists():
                        file_data = json.loads(file.read_text())
                        result.update(file_data)
        return result

    def clear(self) -> None:
        self.result__startup = None
        self.result__teardown = None
        self.result = None

        self.timestamp_last = None
        self.timestamp_start = None

        self.details = {}
        self.exx = None
        self.progress = 0

    @classmethod
    def clear__cls(cls):
        cls.finished__cls = None
        cls.result__startup_cls = None
        cls.result__teardown_cls = None
        for tc in cls.TCS__LIST:
            tc.clear()

    # @classmethod
    # @property
    # def NAME(cls):
    #     return cls.__name__
    #     # return pathlib.Path(__file__).name    # work as last destination where property starts!

    # RESULT ----------------------------------------------------------------------------------------------------------
    @property
    def result(self) -> TYPE__RESULT_W_EXX:
        return self._result

    @result.setter
    def result(self, value: TYPE__RESULT_W_EXX) -> None:
        self._result = value
        self.signals.signal__tc_state_changed.emit(self)

    # # ---------------------------------------------------------
    # @classmethod
    # @property
    # def result__startup_cls(cls) -> Optional[bool]:
    #     return cls.__result__cls_startup
    #
    # @classmethod
    # @result__startup_cls.setter
    # def result__startup_cls(cls, value: Optional[bool]) -> None:
    #     cls.__result__cls_startup = value
    #     # cls.signals.signal__tc_state_changed.emit(cls)

    # DETAILS ---------------------------------------------------------------------------------------------------------
    def details_update(self, details: dict[str, Any]) -> None:
        self.LOGGER.debug("")
        self.details.update(details)
        # self.signals.signal__tc_state_changed.emit(self)

    # =================================================================================================================
    @classmethod
    def run__cls(cls, cls_prev: type[Self] | None = None, single: bool | None = None) -> None | bool:
        """run TC on batch duts(??? may be INDEXES???)
        prefered using in thread on upper level!

        :param cls_prev: use for apply group in sequence
        :param single: if start TC as direct ONE (usually by button)
        :return:
            NONE - if SKIP for any reason
            True - need continue TP
            False - cant continue! need stop TP
        """
        # if not cls.DEVICES__BREEDER_CLS.LIST__DUT:
        #     return

        print(f"run__cls=START={cls.NAME=}={'=' * 50}")

        # GROUP cmp ----------------------------------------------
        if not single and (cls_prev is not None and cls.middle_group__check_equal__cls(cls_prev)):
            cls_prev__equel = True
        else:
            cls_prev__equel = False

        # SKIP ---------------------------------------------------
        if cls.SKIP:
            print(f"run__cls=SKIP={cls.NAME=}={'=' * 50}")
            return
        if cls_prev__equel and not bool(cls_prev.result__startup_cls):
            return

        cls.clear__cls()

        # STARTUP/TERDOWN ----------------------------------------
        if cls_prev__equel:
            cls.result__startup_cls = cls_prev.result__startup_cls
        else:
            if not single and cls_prev:  # and not cls_prev.SKIP:
                cls_prev.teardown__cls()
                if (
                        cls_prev.result__startup_cls is not None
                        and
                        bool(cls_prev.result__startup_cls)
                        and
                        not bool(cls_prev.result__teardown_cls)
                ):  # FIXME: seems need to compare as direct True/Bool
                    return False
            cls.result__startup_cls = cls.startup__cls()

        # WORK ---------------------------------------------------
        if cls.result__startup_cls:
            # BATCH --------------------------
            for tc_inst in cls.TCS__LIST:
                if tc_inst.skip_tc_dut:
                    continue

                print(f"run__cls=tc_inst.start({tc_inst.INDEX=})")
                tc_inst.start()
                if not cls.ASYNC:
                    print(f"run__cls=tc_inst.wait({tc_inst.INDEX=})inONEBYONE")
                    tc_inst.wait()

            # WAIT --------------------------
            if cls.ASYNC:
                for tc_inst in cls.TCS__LIST:
                    print(f"run__cls=tc_inst.wait({tc_inst.INDEX=})inPARALLEL")
                    tc_inst.wait()

        # FINISH -------------------------------------------------
        if single or not cls.result__startup_cls:
            cls.teardown__cls()
        print(f"[TC]FINISH={cls.NAME=}={'=' * 50}")
        return True

    def run(self) -> None:
        self.LOGGER.debug("run")

        # PREPARE --------
        self.clear()
        self.timestamp_start = datetime.datetime.now()
        if (
                not hasattr(self.DEVICES__BREEDER_INST, "DUT")
                or
                self.DEVICES__BREEDER_INST.DUT.SKIP
                or
                not self.DEVICES__BREEDER_INST.DUT.DEV_FOUND
                or
                not self.DEVICES__BREEDER_INST.DUT.connect()
        ):
            return

        # WORK --------
        self.LOGGER.debug("run-startup")
        if self.startup():
            try:
                self.LOGGER.debug("run-run_wrapped START")
                self.result = self.run__wrapped()
                if isinstance(self.result, Valid):
                    self.result.run__if_not_finished()

                self.LOGGER.debug(f"run-run_wrapped FINISHED WITH {self.result=}")
            except Exception as exx:
                self.result = False
                self.exx = exx
        self.LOGGER.debug("run-teardown")
        self.teardown()

    # STARTUP/TEARDOWN ------------------------------------------------------------------------------------------------
    @classmethod
    def startup__cls(cls) -> TYPE__RESULT_W_EXX:
        """before batch work
        """
        print(f"startup__cls")
        cls.finished__cls = False
        # cls.clear__cls()

        result = cls.startup__cls__wrapped
        result = CallableAux(result).resolve_exx()
        if isinstance(result, Valid):
            result.run__if_not_finished()
        print(f"{cls.result__startup_cls=}")
        cls.result__startup_cls = result
        return result

    def startup(self) -> TYPE__RESULT_W_EXX:
        self.LOGGER.debug("")
        self.progress = 1

        result = self.startup__wrapped
        result = CallableAux(result).resolve_exx()
        if isinstance(result, Valid):
            result.run__if_not_finished()
        self.result__startup = result
        return result

    def teardown(self) -> TYPE__RESULT_W_EXX:
        self.LOGGER.debug("")
        self.timestamp_last = datetime.datetime.now()
        self.progress = 99

        result = self.teardown__wrapped
        result = CallableAux(result).resolve_exx()
        if isinstance(result, Valid):
            result.run__if_not_finished()

        self.progress = 100
        self.result__teardown = result
        return result

    @classmethod
    def teardown__cls(cls) -> TYPE__RESULT_W_EXX:
        print(f"run__cls=teardown__cls")

        if not cls.finished__cls or cls.result__teardown_cls is None:
            result = cls.teardown__cls__wrapped
            result = CallableAux(result).resolve_exx()
            if isinstance(result, Valid):
                result.run__if_not_finished()
            cls.result__teardown_cls = result

            if not result:
                print(f"[FAIL]{cls.result__teardown_cls=}//{cls.NAME}")

        else:
            result = cls.result__teardown_cls

        cls.finished__cls = True
        return result

    # =================================================================================================================
    @classmethod
    def terminate__cls(cls) -> None:
        for tc_inst in cls.TCS__LIST:
            try:
                if tc_inst.isRunning() and not tc_inst.isFinished():
                    tc_inst.terminate()
            except:
                pass

        cls.teardown__cls()

    def terminate(self) -> None:
        self.LOGGER.debug("")

        super().terminate()

        progress = self.progress
        self.teardown()
        self.progress = progress

    # REDEFINE ========================================================================================================
    pass
    pass
    pass
    pass
    pass
    pass

    @classmethod
    def startup__cls__wrapped(cls) -> TYPE__RESULT_W_NORETURN:
        return True

    def startup__wrapped(self) -> TYPE__RESULT_W_NORETURN:
        return True

    def run__wrapped(self) -> TYPE__RESULT_W_NORETURN:
        return True

    def teardown__wrapped(self) -> TYPE__RESULT_W_NORETURN:
        return True

    @classmethod
    def teardown__cls__wrapped(cls) -> TYPE__RESULT_W_NORETURN:
        print("HELLO")
        return True


# =====================================================================================================================
class _Info(_TestCaseBase):
    """
    separated class for gen results/info by models!
    """
    INFO_STR__ADD_ATTRS: Iterable[str] = []

    @classmethod
    def get__info__tc(cls) -> dict[str, Any]:
        """
        get info/structure about TcCls
        """
        result = {
            "TC_NAME": cls.NAME,
            "TC_DESCRIPTION": cls.DESCRIPTION,
            "TC_ASYNC": cls.ASYNC,
            "TC_SKIP": cls.SKIP,
            "TC_SETTINGS": cls.settings_read(),
        }
        return result

    # =================================================================================================================
    def get__results_pretty(self) -> str:
        # FIXME: GET FROM INFO_GET????
        result = ""

        result += f"DUT_INDEX={self.INDEX}\n"
        result += f"DUT_SN={self.DEVICES__BREEDER_INST.DUT.SN}\n"
        result += f"DUT_ADDRESS={self.DEVICES__BREEDER_INST.DUT.ADDRESS}\n"
        result += f"tc_skip_dut={self.skip_tc_dut}\n"

        result += f"TC_NAME={self.NAME}\n"
        result += f"TC_GROUP={self.MIDDLE_GROUP__NAME}\n"
        result += f"TC_DESCRIPTION={self.DESCRIPTION}\n"
        result += f"TC_ASYNC={self.ASYNC}\n"
        result += f"TC_SKIP={self.SKIP}\n"

        result += f"SETTINGS=====================\n"
        if self.SETTINGS:
            for name, value in self.SETTINGS.items():
                result += f"{name}:{value}\n"

        result += f"INFO_STR__ADD_ATTRS===========\n"
        if self.INFO_STR__ADD_ATTRS:
            for name in self.INFO_STR__ADD_ATTRS:
                if hasattr(self, name):
                    result += f"{name}:{getattr(self, name)}\n"

        result += f"PROGRESS=====================\n"
        result += f"timestamp_start={self.timestamp_start}\n"
        result += f"timestamp_last={self.timestamp_last}\n"
        result += f"progress={self.progress}\n"
        result += f"exx={self.exx}\n"

        result += "-"*60 + "\n"
        result += f"result__startup={self.result__startup}\n"
        result += "-"*60 + "\n"
        result += f"result={self.result}\n"
        result += "-"*60 + "\n"
        result += f"result__teardown={self.result__teardown}\n"
        result += "-"*60 + "\n"

        result += f"DETAILS=====================\n"
        for name, value in self.details.items():
            result += f"{name}={value}\n"
        return result

    # =================================================================================================================
    def get__results(self, add_info_dut: bool = True, add_info_tc: bool = True) -> dict[str, Any]:
        self.LOGGER.debug("")

        info_dut = {}
        try:
            if add_info_dut:
                info_dut = self.DEVICES__BREEDER_INST.DUT.get__info__dev()
        except:
            pass

        info_tc = {}
        if add_info_tc:
            info_tc = self.get__info__tc()

        result = {
            **info_tc,
            **info_dut,

            # RESULTS
            "timestamp_start": self.timestamp_start,
            "tc_active": self.isRunning(),
            "tc_progress": self.progress,
            "tc_result_startup": bool(self.result__startup),
            "tc_result": None if self.result is None else bool(self.result),
            "tc_details": self.details,
            "result__teardown": bool(self.result__teardown),
            "timestamp_last": self.timestamp_last,
            "log": self.get__results_pretty(),
        }
        return result

    @classmethod
    def get__results__all(cls) -> dict[int, dict[str, Any]]:
        results = {}
        for tc_inst in cls.TCS__LIST:
            results.update({tc_inst.INDEX: tc_inst.get__results()})
        return results


class TestCaseBase(_Info):
    """
    """


# =====================================================================================================================



from typing import *

from base_aux.aux_attr.m1_attr2_nest3_eq_attrs import NestEq_AttrsNotPrivate
from base_aux.aux_attr.m1_attr2_nest4_str import NestStR_Attrs
from base_aux.aux_attr.m1_attr2_nest5_contains import NestContains_AttrIC
from base_aux.aux_attr.m1_attr2_nest6_len import NestLen_AttrNotPrivate

from base_aux.base_inits.m3_nest_init_annots_attrs_by_kwargs import *
from base_aux.aux_attr.m2_annot1_aux import *
from base_aux.aux_attr.m1_attr2_nest3_eq_attrs import *
from base_aux.base_inits.m3_nest_init_annots_attrs_by_kwargs import NestInit_AnnotsAttrByKwArgs


# =====================================================================================================================
class Nest_AttrKit(
    NestInit_AnnotsAttrByKwArgs,    # GAI+Required

    NestEq_AttrsNotPrivate,
    NestStR_Attrs,
    NestContains_AttrIC,
    NestLen_AttrNotPrivate
):     # TODO: decide to delete! use only dynamic?? - NO! keep it!!!
    """
    GOAL
    ----
    just show that child is a kit
    1/ attrs need to init by args/kwargs
    2/ all annotated - must set!

    NOTE
    ----
    !/ DONT USE DIRECTLY! use AttrKit_Blank instead! direct usage acceptable only for isinstance checking!
    1/ used in final CHILDs
    2/ basically used for static values like parsed from ini/json files

    SPECIALLY CREATED FOR
    ---------------------


    OLD docstr
    =======================
        GOAL
        ----
        1/ generate object with exact attrs values by Kwargs like template
        2/ for further comparing by Eq
        3/ all callables will resolve as Exx

        NOTE
        ----
        IgnoreCase applied!

        SAME AS - NestInit_AnnotsAttrByKwArgs
        --------------------------------------
        but
            - args useless - if no annots

        WHY NOT - just EqValid_*
        ------------------------
        1/ cause you will not keep simple direct object with attrs!
        2/ EqValid_* will be created! further!
    """


# =====================================================================================================================
@final      # TODO: decide not use final and use nesting any kit, by collecting attrs???
class AttrKit_Blank(Nest_AttrKit):
    """
    GOAL
    ----
    jast show that you can create any kwargs kit without raising (when check annots required)
    """
    pass


# ---------------------------------------------------------------------------------------------------------------------
@final
class AttrKit_AuthNamePwd(Nest_AttrKit):
    NAME: str
    PWD: str


@final
class AttrKit_AuthTgBot(Nest_AttrKit):
    LINK_ID: str = None     # @mybot20230913
    NAME: str = None        # MyBotPublicName
    TOKEN: str


@final
class AttrKit_AuthServer(Nest_AttrKit):
    NAME: str
    PWD: str
    SERVER: str


# =====================================================================================================================
if __name__ == '__main__':
    pass


# =====================================================================================================================



from typing import *


# =====================================================================================================================
class PROJECT:
    # PROJECT ----------------------------------------------
    NAME_IMPORT: str = "bus_user"
    KEYWORDS: list[str] = [
        "serial", "serial bus", "pyserial", "serial port", "com port", "comport", "rs232", "UART", "TTL",
        "serial client", "serial server", "serial emulator",
        "i2c",
    ]

    # README -----------------------------------------------
    # add DOUBLE SPACE at the end of all lines! for correct representation in MD-viewers
    DESCRIPTION_SHORT: str = "work with equipment over buses like Serial/i2c/... as client and server"
    DESCRIPTION_LONG: str = """
    ###
NOTE: IT SEEMS THIS IS OLD DATA! see tests for actual usage!
    
!. MOST APPROPRIATE COMMAND PROTOCOL
other protocols mot recommended

1. all cmds must be as params (preferred) in equipment or as special command
2. [<CMD_NAME>] - read param value or run special command  
    [IDN] - read value IDN  
    [DUMP] - run special command 
3. [<CMD_NAME> <VALUE>] - write value in parameter or run special cmd with param  
    [VOUT 12.3] - set value into parameter VOUT  
4. [<CMD_NAME> ?] - get available values to write into parameter  
    [MODE ?] - return [0 1 2 3]
5. all command sent must return answer  
    [OK] - if no value was asked
    [<VALUE>] - if asked some value, returned without measurement unit
    [FAIL] - any common not specified error
    [FAIL 0123] - any specified error without description
    [FAIL 02 VALUE OUT OF RANGE] - any specified error with description (full variant)
"""
    FEATURES: list[str] = [
        # "feat1",
        # ["feat2", "block1", "block2"],

        ["[SerialClient]",
            "keep all found ports in base class!",
        ],
        ["Serial",
            "Client+Server",
            "connect with Type__AddressAutoAcceptVariant FIRST_FREE/FIRST_FREE__ANSWER_VALID",
            "set/get params by SlashOrSpacePath addressing",
            "handle BackSpace send manually from terminal",
         ],
        ["SerialServer values",
         "as Callable",
         "ValueUnit",
         "ValueVariants",
         "list_results",
         ],
        ["SerialServer cmd",
         "NONE is equivalent for SUCCESS",
         "no need params (like line_parsed as before)",
         "help - for show all variants (Units/Variants/Callables)!"
         ]
    ]

    # HISTORY -----------------------------------------------
    VERSION: tuple[int, int, int] = (0, 4, 16)
    TODO: list[str] = [
        "add all other port settings into SerialClient",
        "fix all tests! fix EMU",
    ]
    FIXME: list[str] = [
        "add timeout into connect for AndreiK bluetooth comports stack",
    ]
    NEWS: list[str] = [
        "[SerialClient.writeReadLast*] fix timeoutRead on only first read! after first char timeout set default",
    ]


# =====================================================================================================================



from typing import *
from .m2_annot1_aux import AnnotsAux
from base_aux.aux_attr.m1_attr2_nest1_gsai_anycase import *


# =====================================================================================================================
class NestInit_AnnotsRequired(NestGAI_AttrIC):
    """Check all annotated and not defined attributes in instance have values!
    else raise!


    # IS IT STILL USED???
    DONT INHERIT with typing.NamedTuple! will raise!                ???
    For NamedTuple use as separated function with obj parameter!    ???
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        AnnotsAux(self).check_all_defined_or_raise()  # check only after superInit!

    # TODO: deside is it really need NamedTuple and dataclasses??? seems its not need!!! - NEED!!! realise later!!!

    # def annots_get_set(self, obj: Optional[Any] = None) -> set[str]:
    #     """get undefined annotated attributes in class (not instance!)
    #     """
    #     annots = set()
    #     if obj is None:
    #         obj = self
    #     mro = obj.__class__.__mro__
    #     for cls in mro[:-1]:
    #         if cls == tuple and hasattr(obj, "_fields"):    # NamedTuple
    #             return set(obj._fields).difference(set(obj._field_defaults))
    #         if not hasattr(cls, "__annotations__"):     # for last class (tuple) in NamedTuple!
    #             continue
    #         for name in set(cls.__annotations__):
    #             if not hasattr(cls, name):
    #                 annots.update({name, })
    #     # print(annots)
    #     return annots
    #
    # def annots_get_dict(self, obj: Optional[Any] = None) -> Union[TYPING.KWARGS_FINAL, NoReturn]:
    #     """get dict of undefined annotated attributes in class but defined in instance!
    #     """
    #     annots = dict()
    #     for name in self.annots_get_set(obj):
    #         annots.update({name: self.attr_get_ignorecase(name, obj)})
    #     return annots
    #
    # def annots_get_values(self, obj: Optional[Any] = None) -> Union[Iterable[Any], NoReturn]:
    #     """get iterable values of undefined annotated attributes in class but defined in instance!
    #     """
    #     return self.annots_get_dict(obj).values()
    #
    # def attr_get_ignorecase(self, name: str, obj: Optional[Any] = None) -> Union[str, NoReturn]:
    #     """get value for attr name without case sense.
    #     if no attr name in source - raise!
    #     """
    #     if obj is None:
    #         obj = self
    #
    #     attrs_all = list(filter(lambda attr: not callable(getattr(obj, attr)) and not attr.startswith("__"), dir(obj)))
    #
    #     attrs_similar = list(filter(lambda attr: attr.lower() == name.lower(), attrs_all))
    #     if len(attrs_similar) == 1:
    #         return getattr(obj, attrs_similar[0])
    #     elif len(attrs_similar) == 0:
    #         msg = f"[CRITICAL]no[{name=}] in any cases [{attrs_all=}]"
    #         raise Exx__AnnotNotDefined(msg)
    #     else:
    #         msg = f"[CRITICAL]exists several similar [{attrs_similar=}]"
    #         raise Exx__AnnotNotDefined(msg)


# =====================================================================================================================



from base_aux.aux_argskwargs.m2_argskwargs_aux import *
from base_aux.base_statics.m1_types import *


# =====================================================================================================================
class NestInit_SourceKwArgs_Implicite(NestInit_Source):
    """
    NOTE
    ----
    NestInit_SourceKwArgs_Explicite is more useful!

    GOAL
    ----
    just to make inition source with KwArgs
    """
    ARGS: TYPING.ARGS_FINAL
    KWARGS: TYPING.KWARGS_FINAL

    def __init__(self, source: Any = None, *args, **kwargs) -> None:
        self.ARGS = args
        self.KWARGS = kwargs
        super().__init__(source)


# =====================================================================================================================
class NestInit_SourceKwArgs_Explicite(NestInit_Source):
    """
    MORE USEFUL THEN NestInit_SourceKwArgs_Implicite

    GOAL
    ----

    FOR PYTESTAUX!
    """
    ARGS: TYPING.ARGS_FINAL
    KWARGS: TYPING.KWARGS_FINAL

    def __init__(self, source: Any = None, args: TYPING.ARGS_DRAFT = (), kwargs: TYPING.KWARGS_DRAFT = None, *args2, **kwargs2) -> None:
        self.ARGS = ArgsKwargsAux(args).resolve_args()
        self.KWARGS = ArgsKwargsAux(kwargs).resolve_kwargs()
        super().__init__(source, *args2, **kwargs2)


# =====================================================================================================================



from typing import *
from base_aux.buses.m1_serial2_client_derivatives import *


# =====================================================================================================================
class Device(SerialClient_FirstFree_AnswerValid):
    LOG_ENABLE = True
    RAISE_CONNECT = False
    BAUDRATE = 115200
    PREFIX = "ATC:03:"
    EOL__SEND = b"\n"

    REWRITEIF_READNOANSWER = 0
    REWRITEIF_NOVALID = 0

    def address__validate(self) -> bool:
        for _ in range(2):
            if self.write_read__last_validate("get name", "ATC", prefix=self.PREFIX):
                return True

# =====================================================================================================================
if __name__ == "__main__":
    pass

    # emu = Atc_Emulator()
    # emu.start()
    # emu.wait()

    dev = Device()
    print(f"{dev.connect()=}")
    # print(f"{dev.addresses_system__detect()=}")
    print(f"{dev.ADDRESS=}")
    #
    # print(f"{dev.address__validate()=}")
    # print(f"{dev.address__validate()=}")
    # print(f"{dev.address__validate()=}")
    #
    # print(f"{dev.write_read_line_last('get name')=}")
    # print(f"{dev.write_read_line_last('get name')=}")
    # print(f"{dev.write_read_line_last('get name')=}")
    # print(f"{dev.write_read_line_last('get name')=}")
    # print(f"{dev.write_read_line_last('get name')=}")
    # print(f"{dev.disconnect()=}")
    # print(f"{dev.ADDRESS=}")


# =====================================================================================================================



from typing import *
import time

# from base_aux.aux_argskwargs.m1_argskwargs import TYPE__LAMBDA_CONSTRUCTOR
from base_aux.base_statics.m4_enums import *
# from base_aux.aux_types import TypeAux   # CIRCULAR IMPORT
from base_aux.base_inits.m2_nest_init_source_kwargs import NestInit_SourceKwArgs_Implicite

from base_aux.aux_cmp_eq.m2_eq_aux import *


# =====================================================================================================================
class Lambda(NestInit_SourceKwArgs_Implicite):
    """
    IDEA
    ----
    no calling on init!

    GOAL
    ----
    1. (MAIN) delay probable raising on direct func execution (used with NestInit_AttrsLambdaResolve)
    like creating aux_types on Cls attributes
        class Cls:
            ATTR = PrivateValues(123)   # -> Lambda(PrivateValues, 123) - IT IS OLD!!!! but could be used as example!

    2. (not serious) replace simple lambda!
    by using lambda you should define args/kwargs any time! and im sick of it!
        func = lambda *args, **kwargs: sum(*args) + sum(**kwargs.values())  # its not a simple lambda!
        func = lambda *args: sum(*args)  # its simple lambda
        result = func(1, 2)
    replace to
        func = Lambda(sum)
        result = func(1, 2)

        func = Lambda(sum, 1, 2)
        result = func()
    its те a good idea to replace lambda fully!
    cause you cant replace following examples
        func_link = lambda source: str(self.Victim(source))
        func_link = lambda source1, source2: self.Victim(source1) == source2


    SPECIALLY CREATED FOR
    ---------------------
    Item for using with NestInit_AttrsLambdaResolve

    WHY NOT 1=simple LAMBDA?
    ------------------------
    extremely good point!
    but
    1. in case of at least NestInit_AttrsLambdaResolve you cant distinguish method or callable attribute!
    so you explicitly define attributes/aux_types for later constructions
    and in some point it can be more clear REPLACE LAMBDA by this solvation!!!

    2.

    PARAMS
    ======
    :ivar SOURCE: any class or function
    """
    SOURCE: Union[Callable, Any]

    # UNIVERSAL =======================================================================================================
    def construct(self, *args, **kwargs) -> Any | NoReturn:
        """
        unsafe (raise acceptable) get value
        """
        args = args or self.ARGS
        kwargs = kwargs or self.KWARGS
        if callable(self.SOURCE):  # callable accept all variants! TypeAux.check__callable_func_meth_inst_cls!
            return self.SOURCE(*args, **kwargs)
        else:
            return self.SOURCE

    # -----------------------------------------------------------------------------------------------------------------
    def __bool__(self) -> bool | NoReturn:
        return bool(self(*self.ARGS, **self.KWARGS))

    # OVERWRITE! ======================================================================================================
    def __call__(self, *args, **kwargs) -> Any | NoReturn:
        return self.construct(*args, **kwargs)

    def __eq__(self, other) -> bool | NoReturn:
        return EqAux(self()).check_doubleside__bool(other)


# =====================================================================================================================
class LambdaBool(Lambda):
    """
    GOAL
    ----
    same as Lambda, in case of get result in bool variant
    +add reverse

    SPECIALLY CREATED FOR
    ---------------------
    classes.Valid.skip_link with Reverse variant

    why Reversing is so important?
    --------------------------------
    because you cant keep callable link and reversing it by simply NOT
        skip_link__direct = bool        # correct
        skip_link__direct = LambdaBool(bool)  # correct
        skip_link__reversal = not bool  # incorrect
        skip_link__reversal = LambdaBool(bool, attr).get_reverse  # correct

    but here we can use lambda
        skip_link__reversal = lambda attr: not bool(attr)  # correct but not so convenient ???

    PARAMS
    ======
    :ivar BOOL_REVERSE: just for LambdaBoolReversed, no need to init
    """

    BOOL_REVERSE: bool = False

    def __call__(self, *args, **kwargs) -> bool | NoReturn:
        result = bool(self.construct(*args, **kwargs))
        if self.BOOL_REVERSE:
            result = not result
        return result

    def get_reverse(self, *args, **kwargs) -> bool | NoReturn:
        """
        if raise - raise

        try not to use in LambdaBoolReversed
        """
        return not self(*args, **kwargs)

    def get_bool_only(self, *args, **kwargs) -> bool:
        """
        if raise - return False, else get result
        """
        try:
            return self(*args, **kwargs)
        except Exception as exx:
            return False

    def get_bool_only__reverse(self, *args, **kwargs) -> bool:
        return not self.get_bool_only(*args, **kwargs)


class LambdaBoolReversed(LambdaBool):
    """
    just a reversed LambdaBool
    """
    BOOL_REVERSE: bool = True


# =====================================================================================================================
class LambdaTrySuccess(LambdaBool):
    """
    just an ability to check if object is not raised on call

    BEST PRACTICE
    -------------
    1. direct/quick/shortest checks without big trySentence
        if LambdaTrySuccess(func):
            return func()

    2. pytestSkipIf
        @pytest.mark.skipif(LambdaTryFail(func), ...)

    3. pytest assertions

        class Victim(DictAttrAnnotRequired):
            lowercase: str

        assert LambdaTryFail(Victim)
        assert not LambdaTrySuccess(Victim)
        assert LambdaTrySuccess(Victim, lowercase="lowercase")

    EXAMPLES
    --------
        if callables and LambdaTrySuccess(getattr, source, name) and callable(getattr(source, name)):
            continue

        so here raise is acceptable in getattr(source, name) in case of PROPERTY_RAISE
    """
    def __call__(self, *args, **kwargs) -> bool:
        try:
            self.construct(*args, **kwargs)
            return not self.BOOL_REVERSE
        except:
            return bool(self.BOOL_REVERSE)


class LambdaTryFail(LambdaTrySuccess):
    BOOL_REVERSE: bool = True


# =====================================================================================================================
class LambdaSleep(Lambda):
    """
    just delay construction
    """
    WHEN: When2 = When2.BEFORE
    SEC: float = 1

    def __init__(self, sec: float = None, *args, **kwargs) -> None:
        if sec is not None:
            self.SEC = sec
        super().__init__(*args, **kwargs)

    def __call__(self, sec: float = None, *args, **kwargs) -> Any | NoReturn:
        if sec is None:
            sec = self.SEC

        if self.WHEN == When2.BEFORE:
            time.sleep(sec)
        result = self.construct(*args, **kwargs)
        if self.WHEN == When2.AFTER:
            time.sleep(sec)
        return result


# ---------------------------------------------------------------------------------------------------------------------
class LambdaSleepAfter(LambdaSleep):
    """
    CREATED SPECIALLY FOR
    ---------------------
    UART/ATC tests for RST command
    """
    WHEN: When2 = When2.AFTER


# =====================================================================================================================



from base_aux.base_statics.m1_types import *
from base_aux.aux_types.m2_info import *
from base_aux.base_resolver.m1_resolver import *
from base_aux.path1_dir.m1_dirpath import Resolve_DirPath


# =====================================================================================================================
@final
class Resolve_FilePath(NestCall_Resolve):
    """
    GOAL
    ----
    1/ resolve filepath by draft
    2/ combyne by any part
    3/ replace any part

    SPECIALLY CREATED FOR
    ---------------------
    base_aux.files
    """
    NAME: str = ""
    EXTLAST: str = ""
    DIRPATH: TYPING.PATH_FINAL = None

    name_PREFIX: str = ""
    name_SUFFIX: str = ""   # for using in BackUps!

    DOT: bool = None

    # PROPERTIES ------------------------------------------------------------------------------------------------------
    NAMEFINAL: str
    FILEPATH: TYPING.PATH_FINAL

    @property
    def NAMEFULL(self) -> str:
        result = f"{self.name_PREFIX}{self.NAME}{self.name_SUFFIX}"
        if self.DOT:
            result += f"."
        if self.EXTLAST:
            result += f"{self.EXTLAST}"
        return result

    @property
    def FILEPATH(self) -> TYPING.PATH_FINAL:
        return self.DIRPATH.joinpath(self.NAMEFULL)

    # -----------------------------------------------------------------------------------------------------------------
    def __init__(
            self,
            # parts -----
            name: str = None,                   # level1
            extlast: str = None,                # level1
            prefix: str = None,                 # level1
            suffix: str = None,                 # level1

            namefull: str = None,               # level2
            dirpath: TYPING.PATH_DRAFT = None,   # level2

            # full -----
            filepath: TYPING.PATH_DRAFT = None,  # level3
    ):
        """
        NOTE
        ----
        you can use "filepath" as base/default and others (name/extlast/...) for overwrite some of them base parts

        LEVELS param (see in endLine comments)
        --------------------------------------
        1/ params in same level do not affect each other.
        2/ lower level is the major for higher level and will overwrite it.
        see tests for understanding, but it is too obvious when get the idea
        """
        self.apply_filepath(filepath)
        self.apply_dirpath(dirpath or self.DIRPATH)
        self.apply_nameext(namefull)

        # most important! overwrite previous set!
        if prefix is not None:
            self.name_PREFIX = prefix
        if suffix is not None:
            self.name_SUFFIX = suffix
        if name is not None:
            self.NAME = name
        if extlast is not None:
            self.DOT = True
            self.EXTLAST = extlast

    def apply_filepath(self, filepath: TYPING.PATH_DRAFT) -> None:
        if filepath is None:
            return

        filepath = pathlib.Path(filepath)
        self.DIRPATH = filepath.parent
        self.NAME = filepath.stem
        self.EXTLAST = filepath.suffix.rsplit(".", 1)[-1]

        self.DOT = "." in filepath.name

    def apply_dirpath(self, dirpath: TYPING.PATH_DRAFT) -> None:
        self.DIRPATH = Resolve_DirPath(dirpath).resolve()

    def apply_nameext(self, nameext: str) -> None:
        if nameext is None:
            return

        if "." in nameext:
            self.DOT = True

        name_ext: list[str] = nameext.rsplit(".", 1)
        if len(name_ext) == 2:  # DOT exists!
            _name, _extlast = name_ext
            if _name:
                self.NAME = _name
            if _extlast:
                self.EXTLAST = _extlast
        else:
            self.NAME = nameext
            self.EXTLAST = ""

    # -----------------------------------------------------------------------------------------------------------------
    def resolve(self) -> TYPING.PATH_FINAL:
        return self.FILEPATH


# =====================================================================================================================
if __name__ == '__main__':
    obj = pathlib.Path("hello.")
    ObjectInfo(obj).print()


# =====================================================================================================================



from typing import *


# =====================================================================================================================
class PROJECT:
    # AUTHOR -----------------------------------------------
    AUTHOR_NAME: str = "Andrei Starichenko"
    AUTHOR_EMAIL: str = "centroid@mail.ru"
    AUTHOR_HOMEPAGE: str = "https://github.com/centroid457/"

    # PROJECT ----------------------------------------------
    NAME_IMPORT: str = "monitor_utils"
    KEYWORDS: List[str] = [
        "monitor changes", "monitor data",
        "url changes", "url tag changes", "url monitor",
        "imap changes", "email changes", "imap monitor", "email monitor",
        "alerts", "notifications", "email alerts", "telegram alerts",
    ]
    CLASSIFIERS_TOPICS_ADD: List[str] = [
        # "Topic :: Communications",
        # "Topic :: Communications :: Email",
    ]

    # README -----------------------------------------------
    # add DOUBLE SPACE at the end of all lines! for correct representation in MD-viewers
    DESCRIPTION_SHORT: str = "monitor exact data (urlTag/Email) and alert on changes by email/telegram (threading)"
    DESCRIPTION_LONG: str = """
    ## IMPORTANT!
    NOT ALL WEBSITES WORKS! Sportmaster/Acra-rating/...

    ## INSPIRATION
    Suppose you wish to give blood to the Center.
    So nowadays you need to make an appointment by website, BUT you can't do this while the Center actually don't need your group.
    Group necessity shown on Center website and called DonorSvetofor.
    And as result you need monitoring it manually, because there are no news, email notifications, subscriptions.
    It's not difficult but if you do it as day routine (even once a day) its quite distracting.

    So I created it first as Monitor_DonorSvetofor
    """
    FEATURES: List[str] = [
        # "feat1",
        # ["feat2", "block1", "block2"],

        "Threading each monitor",

        ["monitor",
         "website data changes (tag text/attribute)",
         "email received with subject (by regexp) in exact folder", ],

        ["Email/Telegram alert if",
        "monitored data changed (from last state)",
        "html structure was changed so parsing can't be finished",
        "url became unreachable",]
    ]

    # HISTORY -----------------------------------------------
    VERSION: Tuple[int, int, int] = (0, 0, 6)
    TODO: List[str] = [
        "check requirement for python version!",
    ]
    FIXME: List[str] = [
        ["csv is not working on my new python 3.12.1x64!",
            "csv.reader(ofilepath, delimiter=self.CSV_DELIMITER)",
            "module 'private_values.csv' has no attribute 'reader'"
         ]
    ]
    NEWS: List[str] = [
        "[__INIT__.py] fix import",
        "apply last pypi template",
    ]

    # FINALIZE -----------------------------------------------
    NAME_INSTALL: str = NAME_IMPORT.replace("_", "-")


# =====================================================================================================================



from typing import *


# =====================================================================================================================
class PROJECT:
    # AUTHOR -----------------------------------------------
    AUTHOR_NAME: str = "Andrei Starichenko"
    AUTHOR_EMAIL: str = "centroid@mail.ru"
    AUTHOR_HOMEPAGE: str = "https://github.com/centroid457/"

    # PROJECT ----------------------------------------------
    NAME_IMPORT: str = "alerts_msg"
    KEYWORDS: List[str] = [
        "alerts", "notifications",
        "email alerts", "smtp", "mail", "email",
        "telegram alerts", "telegram",
    ]
    CLASSIFIERS_TOPICS_ADD: List[str] = [
        "Topic :: Communications",
        "Topic :: Communications :: Email",
    ]

    # README -----------------------------------------------
    # add DOUBLE SPACE at the end of all lines! for correct representation in MD-viewers
    DESCRIPTION_SHORT: str = "All abilities (mail/telegram) to send alert msgs (threading)"
    DESCRIPTION_LONG: str = """
designed for ...
    """
    FEATURES: List[str] = [
        # "feat1",
        # ["feat2", "block1", "block2"],

        ["send alert msgs", "emails", "telegram", ],
        "threading",
    ]

    # HISTORY -----------------------------------------------
    VERSION: Tuple[int, int, int] = (0, 2, 2)
    TODO: List[str] = [
        "test multyStart by one object (QThread)"
    ]
    FIXME: List[str] = [
        "NEED SOLVE ABILITY to work without PV.FILE!!! as just a http client"
    ]
    NEWS: List[str] = [
        "[__INIT__.py] fix import",
        "apply last pypi template",
    ]

    # FINALIZE -----------------------------------------------
    NAME_INSTALL: str = NAME_IMPORT.replace("_", "-")


# =====================================================================================================================



from base_aux.base_statics.m3_primitives import *
from base_aux.aux_types.m2_info import *


# =====================================================================================================================
class ClsMain(ClsFullTypes):
    attrUpper = "ATTRUPPER"
    attrLower = "attrlower"
    attrSkipFullName = "attrSkipFullName"
    attrSkipPartName = "SkipPartName"


ObjectInfo(
    ClsMain(),
    log_iter=True,
    names__use_only_parts=[],
    names__skip_full=["attrSkipFullName", ],
    names__skip_parts=["SkipPartName", ],
    hide_build_in=None,
).print()
"""
==========================================================================================
----------OBJECTINFO.PRINT--------------------------------------------------------------------------
str(SOURCE)=<__main__.ClsMain object at 0x000001C739A83E00>
repr(SOURCE)=<__main__.ClsMain object at 0x000001C739A83E00>
type(SOURCE)=<class '__main__.ClsMain'>
mro(SOURCE)=['ClsMain', 'ClsFullTypes', 'object']
----------SETTINGS----------------------------------------------------------------------------------
self.NAMES__USE_ONLY_PARTS=[]
self.NAMES__SKIP_FULL=['attrSkipFullName']
self.NAMES__SKIP_PARTS=['init', 'new', 'create', 'enter', 'install', 'set', 'clone', 'copy', 'move', 'next', 'clear', 'reduce', 'close', 'del', 'exit', 'kill', 'exec', 'exec_', 'pyqtConfigure', 'checkout', 'detach', 'run', 'start', 'wait', 'join', 'terminate', 'quit', 'disconnect', 'pop', 'popleft', 'append', 'appendleft', 'extend', 'extendleft', 'add', 'insert', 'reverse', 'rotate', 'sort', 'SkipPartName']
self.HIDE_BUILD_IN=None
self.LOG_ITER=True
self.MAX_LINE_LEN=100
self.MAX_ITER_ITEMS=5
----------log_iter(wait last touched)---------------------------------------------------------------
1:	__class__
2:	__delattr__
3:	__dict__
4:	__dir__
5:	__doc__
6:	__eq__
7:	__format__
8:	__ge__
9:	__getattribute__
10:	__getstate__
11:	__gt__
12:	__hash__
13:	__init__
14:	__init_subclass__
15:	__le__
16:	__lt__
17:	__module__
18:	__ne__
19:	__new__
20:	__reduce__
21:	__reduce_ex__
22:	__repr__
23:	__setattr__
24:	__sizeof__
25:	__str__
26:	__subclasshook__
27:	__weakref__
28:	attrBytes
29:	attrCls
30:	attrClsBoolExx
31:	attrClsBoolFalse
32:	attrClsBoolTrue
33:	attrClsCall
34:	attrClsCallExx
35:	attrClsCallTrue
36:	attrClsGen
37:	attrClsIterYield
38:	attrDict
39:	attrFalse
40:	attrFloat
41:	attrFunc
42:	attrFuncDict
43:	attrFuncExx
44:	attrFuncGen
45:	attrFuncList
46:	attrFuncTrue
47:	attrGenCompr
48:	attrInst
49:	attrInstBooExx
50:	attrInstBoolFalse
51:	attrInstBoolTrue
52:	attrInstCall
53:	attrInstCallExx
54:	attrInstCallTrue
55:	attrInstGen
56:	attrInstIterYield
57:	attrInstMeth
58:	attrInt
59:	attrList
60:	attrListInst
61:	attrLower
62:	attrNone
63:	attrSet
64:	attrSkipFullName
65:	attrSkipPartName
66:	attrStr
67:	attrTrue
68:	attrTuple
69:	attrUpper
70:	attrZero
71:	classmethodNone
72:	methExx
73:	methFunc
74:	methInt
75:	methNone
76:	propertyClassmethodNone
77:	propertyExx
78:	propertyFunc
79:	propertyInt
80:	propertyNone
81:	staticmethodNone
----------SKIPPED_FULLNAMES-------------------------------------------------------------------------
1:	attrSkipFullName
----------SKIPPED_PARTNAMES-------------------------------------------------------------------------
1:	__delattr__
2:	__init__
3:	__init_subclass__
4:	__new__
5:	__reduce__
6:	__reduce_ex__
7:	__setattr__
8:	attrSkipPartName
----------PROPERTIES__ELEMENTARY_SINGLE-------------------------------------------------------------
__doc__             	NoneType    :None
__module__          	str         :__main__
__weakref__         	NoneType    :None
attrBytes           	bytes       :b'bytes'
attrFalse           	bool        :False
attrFloat           	float       :1.1
attrInt             	int         :1
attrLower           	str         :attrlower
attrNone            	NoneType    :None
attrStr             	str         :str
attrTrue            	bool        :True
attrUpper           	str         :ATTRUPPER
attrZero            	int         :0
propertyClassmethodNone	NoneType    :None
propertyInt         	int         :1
propertyNone        	NoneType    :None
----------PROPERTIES__ELEMENTARY_COLLECTION---------------------------------------------------------
__dict__            	dict        :{}
attrDict            	dict        :{1: 1}
attrList            	list        :[1, 2, 3]
attrListInst        	list        :[<pytest_aux.primitives.Cls object at 0x000001C74ACE37D0>, <...
                    	Cls         :	<pytest_aux.primitives.Cls object at 0x000001C74ACE37D0>
                    	Cls         :	<pytest_aux.primitives.Cls object at 0x000001C74ACE37D0>
                    	Cls         :	<pytest_aux.primitives.Cls object at 0x000001C74ACE37D0>
                    	int         :	1
attrSet             	set         :{1, 2, 3}
attrTuple           	tuple       :(1, 2, 3)
----------PROPERTIES__OBJECTS-----------------------------------------------------------------------
attrGenCompr        	generator   :<generator object <genexpr> at 0x000001C74A47D780>
attrInst            	ClsEmpty    :<pytest_aux.primitives.ClsEmpty object at 0x000001C74ACE35F0>
attrInstBooExx      	ClsBoolRaise  :<pytest_aux.primitives.ClsBoolRaise object at 0x000001C74ACE37A0>
attrInstBoolFalse   	ClsBoolFalse:<pytest_aux.primitives.ClsBoolFalse object at 0x000001C74ACE3770>
attrInstBoolTrue    	ClsBoolTrue :<pytest_aux.primitives.ClsBoolTrue object at 0x000001C74ACE3740>
attrInstGen         	ClsGen      :<pytest_aux.primitives.ClsGen object at 0x000001C74ACE3710>
attrInstIterYield   	ClsIterYield:<pytest_aux.primitives.ClsIterYield object at 0x000001C74ACE36E0>
----------PROPERTIES__EXX---------------------------------------------------------------------------
propertyExx         	Exception   :Exception('exxMsg')
----------METHODS__ELEMENTARY_SINGLE----------------------------------------------------------------
__getstate__        	NoneType    :None
__hash__            	int         :122198590432
__repr__            	str         :<__main__.ClsMain object at 0x000001C739A83E00>
__sizeof__          	int         :16
__str__             	str         :<__main__.ClsMain object at 0x000001C739A83E00>
attrFunc            	NoneType    :None
attrFuncTrue        	bool        :True
attrInstCall        	NoneType    :None
attrInstCallTrue    	bool        :True
attrInstMeth        	NoneType    :None
classmethodNone     	NoneType    :None
methInt             	int         :1
methNone            	NoneType    :None
propertyFunc        	NoneType    :None
staticmethodNone    	NoneType    :None
----------METHODS__ELEMENTARY_COLLECTION------------------------------------------------------------
__dir__             	list        :['__module__', 'attrUpper', 'attrLower', 'attrSkipFullName',...
                    	str         :	__module__
                    	str         :	attrUpper
                    	str         :	attrLower
                    	str         :	attrSkipFullName
                    	str         :	attrSkipPartName
                    	            :	...
attrFuncDict        	dict        :{<__main__.ClsMain object at 0x000001C739A83E00>: None}
attrFuncList        	list        :[<__main__.ClsMain object at 0x000001C739A83E00>]
----------METHODS__OBJECTS--------------------------------------------------------------------------
__class__           	ClsMain     :<__main__.ClsMain object at 0x000001C739C51A60>
__subclasshook__    	NotImplementedType:NotImplemented
attrCls             	ClsEmpty    :<pytest_aux.primitives.ClsEmpty object at 0x000001C739E91EB0>
attrClsBoolExx      	ClsBoolRaise  :<pytest_aux.primitives.ClsBoolRaise object at 0x000001C74ACE2690>
attrClsBoolFalse    	ClsBoolFalse:<pytest_aux.primitives.ClsBoolFalse object at 0x000001C74ACE27E0>
attrClsBoolTrue     	ClsBoolTrue :<pytest_aux.primitives.ClsBoolTrue object at 0x000001C74ACE28D0>
attrClsCall         	ClsCall     :<pytest_aux.primitives.ClsCall object at 0x000001C74ACE2990>
attrClsCallExx      	ClsCallExx  :<pytest_aux.primitives.ClsCallExx object at 0x000001C74ACE2A50>
attrClsCallTrue     	ClsCallTrue :<pytest_aux.primitives.ClsCallTrue object at 0x000001C74ACE2AE0>
attrClsGen          	ClsGen      :<pytest_aux.primitives.ClsGen object at 0x000001C74ACE2B40>
attrClsIterYield    	ClsIterYield:<pytest_aux.primitives.ClsIterYield object at 0x000001C74ACE2BA0>
attrFuncGen         	generator   :<generator object FUNC_GEN at 0x000001C74ACCE740>
methFunc            	function    :<function FUNC at 0x000001C74ACEA2A0>
----------METHODS__EXX------------------------------------------------------------------------------
__eq__              	TypeError   :TypeError('expected 1 argument, got 0')
__format__          	TypeError   :TypeError('ClsMain.__format__() takes exactly one argument (...
__ge__              	TypeError   :TypeError('expected 1 argument, got 0')
__getattribute__    	TypeError   :TypeError('expected 1 argument, got 0')
__gt__              	TypeError   :TypeError('expected 1 argument, got 0')
__le__              	TypeError   :TypeError('expected 1 argument, got 0')
__lt__              	TypeError   :TypeError('expected 1 argument, got 0')
__ne__              	TypeError   :TypeError('expected 1 argument, got 0')
attrFuncExx         	Exception   :Exception('CALLABLE_EXX')
attrInstCallExx     	Exception   :Exception()
methExx             	Exception   :Exception('exxMsg')
==========================================================================================
"""

ObjectInfo(
    ClsMain(),
    # log_iter=False,
    names__use_only_parts="attr",
    # names__skip_full=["attrSkipFullName", ],
    # names__skip_parts=["SkipPartName", ],
    # hide_build_in=None,
    # max_line_len=0,
    # max_iter_items=0,
).print()
"""
==========================================================================================
----------OBJECTINFO.PRINT--------------------------------------------------------------------------
str(SOURCE)=<__main__.ClsMain object at 0x000001C748B236B0>
repr(SOURCE)=<__main__.ClsMain object at 0x000001C748B236B0>
type(SOURCE)=<class '__main__.ClsMain'>
mro(SOURCE)=['ClsMain', 'ClsFullTypes', 'object']
----------SETTINGS----------------------------------------------------------------------------------
self.NAMES__USE_ONLY_PARTS=['attr']
self.NAMES__SKIP_FULL=['attrSkipFullName']
self.NAMES__SKIP_PARTS=['init', 'new', 'create', 'enter', 'install', 'set', 'clone', 'copy', 'move', 'next', 'clear', 'reduce', 'close', 'del', 'exit', 'kill', 'exec', 'exec_', 'pyqtConfigure', 'checkout', 'detach', 'run', 'start', 'wait', 'join', 'terminate', 'quit', 'disconnect', 'pop', 'popleft', 'append', 'appendleft', 'extend', 'extendleft', 'add', 'insert', 'reverse', 'rotate', 'sort', 'SkipPartName']
self.HIDE_BUILD_IN=None
self.LOG_ITER=None
self.MAX_LINE_LEN=100
self.MAX_ITER_ITEMS=5
----------log_iter(wait last touched)---------------------------------------------------------------
----------SKIPPED_FULLNAMES-------------------------------------------------------------------------
1:	attrSkipFullName
----------SKIPPED_PARTNAMES-------------------------------------------------------------------------
1:	__delattr__
2:	__setattr__
3:	attrSkipPartName
----------PROPERTIES__ELEMENTARY_SINGLE-------------------------------------------------------------
attrBytes           	bytes       :b'bytes'
attrFalse           	bool        :False
attrFloat           	float       :1.1
attrInt             	int         :1
attrLower           	str         :attrlower
attrNone            	NoneType    :None
attrStr             	str         :str
attrTrue            	bool        :True
attrUpper           	str         :ATTRUPPER
attrZero            	int         :0
----------PROPERTIES__ELEMENTARY_COLLECTION---------------------------------------------------------
attrDict            	dict        :{1: 1}
attrList            	list        :[1, 2, 3]
attrListInst        	list        :[<pytest_aux.primitives.Cls object at 0x000001C74ACE37D0>, <...
                    	Cls         :	<pytest_aux.primitives.Cls object at 0x000001C74ACE37D0>
                    	Cls         :	<pytest_aux.primitives.Cls object at 0x000001C74ACE37D0>
                    	Cls         :	<pytest_aux.primitives.Cls object at 0x000001C74ACE37D0>
                    	int         :	1
attrSet             	set         :{1, 2, 3}
attrTuple           	tuple       :(1, 2, 3)
----------PROPERTIES__OBJECTS-----------------------------------------------------------------------
attrGenCompr        	generator   :<generator object <genexpr> at 0x000001C74A47D780>
attrInst            	ClsEmpty    :<pytest_aux.primitives.ClsEmpty object at 0x000001C74ACE35F0>
attrInstBooExx      	ClsBoolRaise  :<pytest_aux.primitives.ClsBoolRaise object at 0x000001C74ACE37A0>
attrInstBoolFalse   	ClsBoolFalse:<pytest_aux.primitives.ClsBoolFalse object at 0x000001C74ACE3770>
attrInstBoolTrue    	ClsBoolTrue :<pytest_aux.primitives.ClsBoolTrue object at 0x000001C74ACE3740>
attrInstGen         	ClsGen      :<pytest_aux.primitives.ClsGen object at 0x000001C74ACE3710>
attrInstIterYield   	ClsIterYield:<pytest_aux.primitives.ClsIterYield object at 0x000001C74ACE36E0>
----------PROPERTIES__EXX---------------------------------------------------------------------------
----------METHODS__ELEMENTARY_SINGLE----------------------------------------------------------------
attrFunc            	NoneType    :None
attrFuncTrue        	bool        :True
attrInstCall        	NoneType    :None
attrInstCallTrue    	bool        :True
attrInstMeth        	NoneType    :None
----------METHODS__ELEMENTARY_COLLECTION------------------------------------------------------------
attrFuncDict        	dict        :{<__main__.ClsMain object at 0x000001C748B236B0>: None}
attrFuncList        	list        :[<__main__.ClsMain object at 0x000001C748B236B0>]
----------METHODS__OBJECTS--------------------------------------------------------------------------
attrCls             	ClsEmpty    :<pytest_aux.primitives.ClsEmpty object at 0x000001C74ACE39E0>
attrClsBoolExx      	ClsBoolRaise  :<pytest_aux.primitives.ClsBoolRaise object at 0x000001C74ACE3A10>
attrClsBoolFalse    	ClsBoolFalse:<pytest_aux.primitives.ClsBoolFalse object at 0x000001C74ACE3A70>
attrClsBoolTrue     	ClsBoolTrue :<pytest_aux.primitives.ClsBoolTrue object at 0x000001C74ACE3AA0>
attrClsCall         	ClsCall     :<pytest_aux.primitives.ClsCall object at 0x000001C74ACE3AD0>
attrClsCallExx      	ClsCallExx  :<pytest_aux.primitives.ClsCallExx object at 0x000001C74ACE3B00>
attrClsCallTrue     	ClsCallTrue :<pytest_aux.primitives.ClsCallTrue object at 0x000001C74ACE3B30>
attrClsGen          	ClsGen      :<pytest_aux.primitives.ClsGen object at 0x000001C74ACE3B60>
attrClsIterYield    	ClsIterYield:<pytest_aux.primitives.ClsIterYield object at 0x000001C74ACE3B90>
attrFuncGen         	generator   :<generator object FUNC_GEN at 0x000001C74ACCE8E0>
----------METHODS__EXX------------------------------------------------------------------------------
__getattribute__    	TypeError   :TypeError('expected 1 argument, got 0')
attrFuncExx         	Exception   :Exception('CALLABLE_EXX')
attrInstCallExx     	Exception   :Exception()
==========================================================================================
"""



from typing import *
from base_aux.aux_iter.m1_iter_aux import IterAux
from base_aux.aux_values.m0_novalue import *


# =====================================================================================================================
class DictIgnorecase(dict):
    """
    just a Caseinsense dict
    """
    # __getattr__ = dict.get
    # __setattr__ = dict.__setitem__
    # __delattr__ = dict.__delitem__
    # __iter__ = dict.__iter__
    # __copy__ = dict.copy

    # __repr__ = dict.__repr__  # и так работает!
    # __str__ = dict.__str__    # и так работает!
    # __len__ = dict.__len__    # и так работает!

    # -----------------------------------------------------------------------------------------------------------------
    # GENERIC CLASSES LIKE DICT MUST APPLIED LAST IN MRO!!!
    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)

    # -----------------------------------------------------------------------------------------------------------------
    def pop(self, item: Any) -> None:
        item_original = IterAux(self).item__get_original(item)
        if item_original is NoValue:
            item_original = item

        super().pop(item_original)

    def get(self, item: Any) -> Any:    # | NoReturn:
        """
        always get value or None!
        if you need check real contain key - check contain)))) [assert key in self] or [getitem_original]
        """
        key_original = IterAux(self).item__get_original(item)
        if key_original is NoValue:
            return None
            # key_original = item
            # msg = f"{item=}"
            # raise KeyError(msg)

        return super().get(key_original)

    # def set(self, item: Any, value: Any) -> None:

    def update(self, m, /, **kwargs) -> None:
        for item, value in m.items():
            key_original = IterAux(self).item__get_original(item)
            if key_original is NoValue:
                key_original = item

            super().update({key_original: value})

    def __contains__(self, item: Any) -> bool:
        if IterAux(self).item__get_original(item) is not NoValue:
            return True

    # -----------------------------------------------------------------------------------------------------------------
    # ITEM is universal!
    def __getitem__(self, item: Any) -> Any | NoReturn:
        return self.get(item)

    def __setitem__(self, item: Any, value: Any) -> None | NoReturn:
        self.update({item: value})

    def __delitem__(self, item: Any) -> None:
        self.pop(item)


# =====================================================================================================================



"""
GAME
----

https://wordleplay.com/ru/octordle
https://wordleplay.com/ru/wordle-solver

CODE REALIZASIONS
-----------------
https://habr.com/ru/articles/818883/


WORDS source
------------
https://github.com/cwaterloo/5bukv/blob/main/russian5.txt   -5letters for exact game
https://gist.github.com/kissarat/bd30c324439cee668f0ac76732d6c825   -all counts (notFull!!!)

NOTE
----
how to find new words

1/ MANUALLY
find combinations
split groupes by first letter
check by this code
"""

import pathlib
from base_aux.aux_text.m1_text_aux import TextAux
from base_aux.base_inits.m2_nest_init_source_kwargs import *
from base_aux.base_inits.m4_nest_init_annots_by_types import *


# =====================================================================================================================
def check_lack_words() -> None:
    applicants: list[str] = """
проверка_СТАРТ 

ЯГОДА
ЯГУАР

проверка_ФИНИШ
    """.lower().split()

    file = pathlib.Path(__file__, "..", "game1_nouns5rus.txt")
    text = file.read_text(encoding="utf8").lower()

    words: set[str] = set(TextAux(text).split_lines(True))

    for item in applicants:
        if item and item not in words:
            print(item)


# =====================================================================================================================
@final
class CharMask(NestInit_AnnotsByTypes_NotExisted):
    # HIDDEN: str
    # ATTEMPTS: list[str]
    POS_DET: list[str]
    POS_EXCL: list[set[str]]
    INCL: set[str]
    EXCL: set[str]

    def __init__(self, length: int):
        super().__init__()
        self.POS_DET = ["", ] * length
        self.POS_EXCL = [set(), ] * length

    @property
    def POS_DET_WM(self) -> str:
        result = ""
        for pos in self.POS_DET:
            if not pos:
                pos = "?"
            result += pos
        return result

    def __str__(self):
        INCL = f"[{''.join(self.INCL)}]"
        EXCL = f"[{''.join(self.EXCL)}]"
        POS_EXCL = str(self.POS_EXCL).replace(" ", "")
        return f"{self.__class__.__name__}({self.POS_DET_WM},{POS_EXCL},{INCL=},{EXCL=})"


# ---------------------------------------------------------------------------------------------------------------------
class FilterMask(NestInit_SourceKwArgs_Implicite):
    SOURCE: str
    ARGS: tuple[str, ...]    # ATTEMPTS

    CHARMASK: CharMask

    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)
    #     self.get_mask()

    def get_mask(self) -> CharMask:
        self.CHARMASK = CharMask(5)
        # self.CHARMASK.HIDDEN = self.SOURCE
        # result = CharMask(5)
        for attempt in self.ARGS:
            self.attemtp_apply(attempt)

        return self.CHARMASK

    def attemtp_apply(self, attempt: str) -> None:
        for index, char_i in enumerate(attempt):
            if char_i in self.SOURCE:
                self.CHARMASK.INCL.update(char_i)

                if self.SOURCE[index] == char_i:
                    self.CHARMASK.POS_DET[index] = char_i
                else:
                    self.CHARMASK.POS_EXCL[index].update(char_i)
            else:
                self.CHARMASK.EXCL.update(char_i)


# ---------------------------------------------------------------------------------------------------------------------
def check_mask() -> None:
    pass


# =====================================================================================================================
if __name__ == "__main__":
    check_lack_words()


# =====================================================================================================================



from typing import *
import numpy as np

from base_aux.base_inits.m1_nest_init_source import *


# =====================================================================================================================
@final
class ArrayAux(NestInit_Source):
    SOURCE: np.ndarray

    def d2_get_compact_str(
        self,
        interpreter: Optional[dict[Any, Any]] = None,
        separate_rows: Optional[int] = None,
        wrap: Optional[bool] = None,
        use_rows_num: Optional[bool] = None
    ) -> str:
        """

        :param array:
        :param interpreter: dictionary to change some elements
        :param separate_rows: add blank line on step
        :param wrap: add additional strings before and after data
        :param use_rows_num: add row num in
        :return:
        """
        interpreter = interpreter or {}
        count_rows = len(self.SOURCE)
        count_columns = len(self.SOURCE[0])
        row_pos = 0
        result: str = ""

        # tab_row_nums = (count_rows +1)//10 + 1
        # if (count_rows+1)%10 > 0:
        #     tab_row_nums += 1

        tab_row_nums = 4

        if wrap:
            if use_rows_num:
                result += " " * tab_row_nums
            result += "=" * count_columns + "\n"

        for row in self.SOURCE:
            row_pos += 1
            if separate_rows and row_pos > 1 and (row_pos - 1) % separate_rows == 0:
                result += f"\n"

            if use_rows_num:
                result += "{:{width}}".format(str(row_pos), width=tab_row_nums)
            for value in row:
                replaced = interpreter.get(value)
                if replaced is not None:
                    value = replaced
                result += f"{value}"

            if row_pos != count_rows:
                result += f"\n"

        if wrap:
            result += "\n"
            if use_rows_num:
                result += " " * tab_row_nums
            result += "=" * count_columns

        return result


# =====================================================================================================================
